#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for schooltool views.
"""

import unittest
import doctest

from zope.interface import classImplements
from zope.component import provideAdapter
from zope.annotation.interfaces import IAttributeAnnotatable

from schooltool.app.browser.testing import setUp as browserSetUp, tearDown
from schooltool.testing import setup


def setUp(test=None):
    browserSetUp(test)
    setup.setUpCalendaring()

    from schooltool.app.overlay import CalendarOverlayInfo
    from schooltool.app.app import ShowTimetables
    provideAdapter(ShowTimetables)
    classImplements(CalendarOverlayInfo, IAttributeAnnotatable)


# def doctest_CalendarOverlayView():
#     r"""Tests for CalendarOverlayView

#     Some initial setup:

#         >>> from schooltool.app.browser.overlay import CalendarOverlayView
#         >>> View = SimpleViewClass('../templates/calendar_overlay.pt',
#         ...                        bases=(CalendarOverlayView,))

#     CalendarOverlayView is a view on anything.

#         >>> context = object()
#         >>> request = TestRequest()
#         >>> view = View(context, request, None, None)

#     It renders to an empty string unless its context is the calendar of the
#     authenticated user

#         >>> view()
#         u'\n'

#     If you are an authenticated user looking at your own calendar, this view
#     renders a calendar selection portlet.

#         >>> from schooltool.group.group import Group
#         >>> from schooltool.person.person import Person
#         >>> from schooltool.app.security import Principal
#         >>> app = setup.setUpSchoolToolSite()
#         >>> person = app['persons']['whatever'] = Person('fred')
#         >>> group1 = app['groups']['g1'] = Group(title="Group 1")
#         >>> group2 = app['groups']['g2'] = Group(title="Group 2")
#         >>> person.overlaid_calendars.add(ISchoolToolCalendar(group1))
#         <...CalendarOverlayInfo object at ...>
#         >>> person.overlaid_calendars.add(ISchoolToolCalendar(group2),
#         ...                               show=False)
#         <...CalendarOverlayInfo object at ...>

#         >>> request = TestRequest()
#         >>> request.setPrincipal(Principal('id', 'title', person))
#         >>> view = View(ISchoolToolCalendar(person), request, None, None)

#         >>> print view()
#         <div id="portlet-calendar-overlay" class="portlet">
#         ...
#         <td><input type="checkbox" name="overlay:list"
#                checked="checked" value="/groups/g1" /></td>
#         ...
#         <td><input type="checkbox" name="overlay:list"
#                value="/groups/g2" /></td>
#         ...
#         </div>

#     If the request has 'OVERLAY_APPLY', CalendarOverlayView applies your
#     changes

#         >>> request.form['overlay'] = [u'/groups/g2']
#         >>> request.form['OVERLAY_APPLY'] = u"Apply"
#         >>> print view()
#         <div id="portlet-calendar-overlay" class="portlet">
#         ...
#         <td><input type="checkbox" name="overlay:list"
#                value="/groups/g1" /></td>
#         ...
#         <td><input type="checkbox" name="overlay:list"
#                checked="checked" value="/groups/g2" /></td>
#         ...
#         </div>

#     It also redirects you to request.URL:

#         >>> request.response.getStatus()
#         302
#         >>> request.response.getHeader('Location')
#         'http://127.0.0.1'

#     There are two reasons for the redirect: first, part of the page template
#     just rendered might have become invalid when calendar overlay selection
#     changed, second, this lets the user refresh the page without having to
#     experience confirmations dialogs that say "Do you want to POST this form
#     again?".

#     If the request has 'OVERLAY_MORE', CalendarOverlayView redirects to
#     calendar_selection.html

#         >>> request = TestRequest()
#         >>> request.setPrincipal(Principal('id', 'title', person))
#         >>> request.form['OVERLAY_MORE'] = u"More..."
#         >>> view = View(ISchoolToolCalendar(person), request, None, None)
#         >>> content = view()
#         >>> request.response.getStatus()
#         302
#         >>> request.response.getHeader('Location')
#         'http://127.0.0.1/persons/fred/calendar_selection.html?nexturl=http%3A//127.0.0.1'

#     """

# def doctest_CalendarOverlayView_items():
#     """Tests for CalendarOverlayView.items().

#         >>> from schooltool.app.browser.overlay import CalendarOverlayView

#     We will need some persons and groups for the demonstration.

#         >>> from schooltool.group.group import Group
#         >>> from schooltool.person.person import Person
#         >>> app = setup.setUpSchoolToolSite()
#         >>> person = app['persons']['p1'] = Person('p1', title="Person")
#         >>> group1 = app['groups']['g1'] = Group(title="Group 1")
#         >>> group2 = app['groups']['g2'] = Group(title="Group 2")

#     When the person has no calendars in his overlay list, items returns
#     an empty list

#         >>> from zope.publisher.browser import TestRequest
#         >>> from schooltool.app.security import Principal
#         >>> request = TestRequest()
#         >>> request.setPrincipal(Principal('', '', person))
#         >>> context = ISchoolToolCalendar(person)
#         >>> view = CalendarOverlayView(context, request, None, None)
#         >>> view.items()
#         []

#     When the person has calendars in his overlay list

#         >>> person.overlaid_calendars.add(ISchoolToolCalendar(group2))
#         <...CalendarOverlayInfo object at ...>
#         >>> person.overlaid_calendars.add(ISchoolToolCalendar(group1),
#         ...                               show=False)
#         <...CalendarOverlayInfo object at ...>

#         >>> from zope.testing.doctestunit import pprint
#         >>> pprint(view.items())
#         [{'calendar': <schooltool.app.cal.Calendar object at ...>,
#           'checked': '',
#           'color1': '#eed680',
#           'color2': '#d1940c',
#           'id': u'/groups/g1',
#           'title': 'Group 1'},
#          {'calendar': <schooltool.app.cal.Calendar object at ...>,
#           'checked': 'checked',
#           'color1': '#e0b6af',
#           'color2': '#c1665a',
#           'id': u'/groups/g2',
#           'title': 'Group 2'}]

#     """


# def doctest_CalendarOverlayView_items_with_identical_titles():
#     """Tests for CalendarOverlayView.items().

#         >>> from schooltool.app.browser.overlay import CalendarOverlayView

#     We will need some persons and groups for the demonstration.

#         >>> from schooltool.group.group import Group
#         >>> from schooltool.person.person import Person
#         >>> app = setup.setUpSchoolToolSite()
#         >>> person = app['persons']['p1'] = Person('p1', title="Person")
#         >>> group1 = app['groups']['g1'] = Group(title="Group")
#         >>> group5 = app['groups']['g5'] = Group(title="Group")
#         >>> group3 = app['groups']['g3'] = Group(title="Group")
#         >>> group4 = app['groups']['g4'] = Group(title="Group")
#         >>> group2 = app['groups']['g2'] = Group(title="Group")

#     When the person has calendars in his overlay list

#         >>> from zope.publisher.browser import TestRequest
#         >>> from schooltool.app.security import Principal
#         >>> request = TestRequest()
#         >>> request.setPrincipal(Principal('', '', person))
#         >>> context = ISchoolToolCalendar(person)

#         >>> for group in app['groups'].values():
#         ...     link = person.overlaid_calendars.add(ISchoolToolCalendar(group))

#         >>> view = CalendarOverlayView(context, request, None, None)
#         >>> from zope.testing.doctestunit import pprint
#         >>> pprint(view.items())
#         [{'calendar': <schooltool.app.cal.Calendar object at ...>,
#           'checked': 'checked',
#           'color1': '#e0b6af',
#           'color2': '#c1665a',
#           'id': u'/groups/g1',
#           'title': 'Group'},
#          {'calendar': <schooltool.app.cal.Calendar object at ...>,
#           'checked': 'checked',
#           'color1': '#eed680',
#           'color2': '#d1940c',
#           'id': u'/groups/g2',
#           'title': 'Group'},
#          {'calendar': <schooltool.app.cal.Calendar object at ...>,
#           'checked': 'checked',
#           'color1': '#c5d2c8',
#           'color2': '#83a67f',
#           'id': u'/groups/g3',
#           'title': 'Group'},
#          {'calendar': <schooltool.app.cal.Calendar object at ...>,
#           'checked': 'checked',
#           'color1': '#efe0cd',
#           'color2': '#e0c39e',
#           'id': u'/groups/g4',
#           'title': 'Group'},
#          {'calendar': <schooltool.app.cal.Calendar object at ...>,
#           'checked': 'checked',
#           'color1': '#ada7c8',
#           'color2': '#887fa3',
#           'id': u'/groups/g5',
#           'title': 'Group'}]

#     """


# def doctest_CalendarSelectionView():
#     """Tests for CalendarSelectionView

#         >>> from schooltool.app.interfaces import ISchoolToolApplication
#         >>> from schooltool.app.interfaces import IApplicationPreferences
#         >>> from schooltool.app.app import getApplicationPreferences
#         >>> provideAdapter(getApplicationPreferences,
#         ...                (ISchoolToolApplication,), IApplicationPreferences)
#         >>> from schooltool.app.browser.overlay import CalendarSelectionView
#         >>> View = SimpleViewClass('../templates/calendar_selection.pt',
#         ...                        bases=(CalendarSelectionView,))

#     CalendarSelectionView is a view on IPerson

#         >>> from schooltool.resource.resource import Resource
#         >>> from schooltool.group.group import Group
#         >>> from schooltool.person.person import Person
#         >>> from schooltool.app.security import Principal
#         >>> app = setup.setUpSchoolToolSite()
#         >>> persons = app['persons']
#         >>> groups = app['groups']
#         >>> resources = app['resources']
#         >>> fred = persons['fred'] = Person('fred', 'Fred F.')
#         >>> eric = persons['eric'] = Person('eric', 'Eric Bjornsen')
#         >>> igor = persons['igor'] = Person('igor', 'Igor')
#         >>> admins = groups['admins'] = Group('Administrators')
#         >>> car = resources['car'] = Resource('Company car')
#         >>> request = TestRequest()
#         >>> request.setPrincipal(Principal('fred', '', fred))
#         >>> view = View(fred, request)

#     It lists Eric's, Igor's calendars and the site-wide calendar as available
#     for selection

#         >>> print view()
#         <BLANKLINE>
#         ...
#             Select calendars to display
#         ...
#               <legend>Public Calendar</legend>
#               <label>
#                 <input type="checkbox" name="application"
#                        value="application" />
#         ...
#             <fieldset class="inline">
#               <legend>People</legend>
#               <select size="8" multiple="multiple" id="people"
#                       name="persons:list">
#                 <option value="eric">Eric Bjornsen</option>
#                 <option value="igor">Igor</option>
#               </select>
#             </fieldset>
#         <BLANKLINE>
#             <fieldset class="inline">
#               <legend>Groups</legend>
#               <select size="8" multiple="multiple" id="groups"
#                       name="groups:list">
#                 <option value="admins">Administrators</option>
#               </select>
#             </fieldset>
#         <BLANKLINE>
#             <fieldset class="inline">
#               <legend>Resources</legend>
#               <select size="8" multiple="multiple" id="resources"
#                       name="resources:list">
#                 <option value="car">Company car</option>
#               </select>
#             </fieldset>
#         ...

#     If a person's calendar is added to your overlaid calendars list, you
#     can see that in the form.

#         >>> fred.overlaid_calendars.add(ISchoolToolCalendar(eric), show=False)
#         <...CalendarOverlayInfo object at ...>

#         >>> print view()
#         <BLANKLINE>
#         ...
#               <select size="8" multiple="multiple" id="people"
#                       name="persons:list">
#                 <option selected="selected" value="eric">Eric Bjornsen</option>
#                 <option value="igor">Igor</option>
#               </select>
#         ...

#     Note that the user does not see his own calendar in that list:

#         >>> 'value="fred"' in view()
#         False

#     We can submit that form

#         >>> request.form["persons"] = [u"eric", u"igor"]
#         >>> request.form["groups"] = [u"admins"]
#         >>> request.form["resources"] = [u"car"]
#         >>> request.form["UPDATE_SUBMIT"] = u"Apply"
#         >>> print view()
#         <BLANKLINE>
#         ...
#               <select size="8" multiple="multiple" id="people"
#                       name="persons:list">
#                 <option selected="selected" value="eric">Eric Bjornsen</option>
#                 <option selected="selected" value="igor">Igor</option>
#               </select>
#         ...

#     We can see that the calendars we selected were added to the list

#         >>> ISchoolToolCalendar(igor) in fred.overlaid_calendars
#         True
#         >>> ISchoolToolCalendar(admins) in fred.overlaid_calendars
#         True
#         >>> ISchoolToolCalendar(car) in fred.overlaid_calendars
#         True

#     We can also remove calendars

#         >>> request.form["persons"] = [u"igor"]
#         >>> request.form["UPDATE_SUBMIT"] = u"Apply"
#         >>> print view()
#         <BLANKLINE>
#         ...
#               <select size="8" multiple="multiple" id="people"
#                       name="persons:list">
#                 <option value="eric">Eric Bjornsen</option>
#                 <option selected="selected" value="igor">Igor</option>
#               </select>
#         ...

#         >>> ISchoolToolCalendar(eric) in fred.overlaid_calendars
#         False

#     When you submit the form, you are redirected back to the original view

#         >>> request.form["nexturl"] = 'http://localhost/persons/fred/calendar'
#         >>> request.form["UPDATE_SUBMIT"] = u"Apply"
#         >>> output = view()
#         >>> request.response.getStatus()
#         302
#         >>> request.response.getHeader('Location')
#         'http://localhost/persons/fred/calendar'

#     The same thing happens if you press Cancel:

#         >>> request = TestRequest()
#         >>> request.form["nexturl"] = 'http://localhost/persons/fred/calendar'
#         >>> request.form["CANCEL"] = u"Cancel"
#         >>> view = View(fred, request)
#         >>> output = view()
#         >>> request.response.getStatus()
#         302
#         >>> request.response.getHeader('Location')
#         'http://localhost/persons/fred/calendar'

#     Regression test for issue328 (traceback when adding overlays).  This was
#     caused by getApplicationCalendar returning None if it could not access the
#     application calendar.  To solve this we now return an empty dict and check
#     the result in update()

#         >>> view.getApplicationCalendar()
#         {}
#         >>> view.update()

#     """


# def doctest_CalendarSelectionView_updateSelection():
#     """Test for CalendarSelectionView._updateSelection

#         >>> from schooltool.app.browser.overlay import CalendarSelectionView
#         >>> from schooltool.app.interfaces import IShowTimetables
#         >>> from schooltool.person.person import Person
#         >>> from schooltool.group.group import Group
#         >>> from schooltool.resource.resource import Resource
#         >>> from schooltool.app.cal import Calendar
#         >>> request = TestRequest()
#         >>> context = None
#         >>> view = CalendarSelectionView(context, request)
#         >>> user = Person()
#         >>> book = Resource('A book')
#         >>> beamer = Resource('A beamer')
#         >>> friends = Group('Friends')
#         >>> joe = Person(title='Joe')
#         >>> appcal = Calendar(Resource('Application'))

#         >>> def stub_getCalendars(container):
#         ...     if container == 'persons':
#         ...         return [{'id': 'joe',
#         ...                  'title': 'Joe',
#         ...                  'selected': ISchoolToolCalendar(joe)
#         ...                                 in user.overlaid_calendars,
#         ...                  'calendar': ISchoolToolCalendar(joe)}]
#         ...     elif container == 'groups':
#         ...         return [{'id': 'friends',
#         ...                  'title': 'Friends',
#         ...                  'selected': ISchoolToolCalendar(friends)
#         ...                                 in user.overlaid_calendars,
#         ...                  'calendar': ISchoolToolCalendar(friends)}]
#         ...     elif container == 'resources':
#         ...         return [{'id': 'book',
#         ...                  'title': 'The book',
#         ...                  'selected': ISchoolToolCalendar(book)
#         ...                                 in user.overlaid_calendars,
#         ...                  'calendar': ISchoolToolCalendar(book)},
#         ...                 {'id': 'beamer',
#         ...                  'title': 'A beamer',
#         ...                  'selected': ISchoolToolCalendar(beamer)
#         ...                                 in user.overlaid_calendars,
#         ...                  'calendar': ISchoolToolCalendar(beamer)}]
#         ...     else:
#         ...         return []
#         >>> view.getCalendars = stub_getCalendars
#         >>> def stub_getApplicationCalendar():
#         ...     return {'title': 'Applcation',
#         ...             'selected': appcal in user.overlaid_calendars,
#         ...             'calendar': appcal}
#         >>> view.getApplicationCalendar = stub_getApplicationCalendar

#         >>> def print_overlays(user=user):
#         ...     l = [(cal.calendar.title, cal)
#         ...          for cal in user.overlaid_calendars]
#         ...     l.sort()
#         ...     for title, info in l:
#         ...         showtt = IShowTimetables(info).showTimetables
#         ...         print '[%s][%s] %s' % (info.show and '+' or '-',
#         ...                                showtt and '+' or '-',
#         ...                                title)

#     So, we have two calendars available for selection, but we choose none of
#     them.  Nothing happens.

#         >>> view._updateSelection(user)
#         >>> print_overlays()

#     We can choose one, and it appears in the overlay list.

#         >>> request.form['resources'] = ['beamer']
#         >>> view._updateSelection(user)
#         >>> print_overlays()
#         [+][-] A beamer

#     Note that the calendar overlay is visible by default (indicated by '+' in
#     the first checkbox), but the timetable is hidden (indicated by '-' in the
#     second checkbox).

#     We can choose both

#         >>> request.form['resources'] = ['beamer', 'book']
#         >>> view._updateSelection(user)
#         >>> print_overlays()
#         [+][-] A beamer
#         [+][-] A book

#     We can choose one again, and the other disappears

#         >>> request.form['resources'] = ['book']
#         >>> view._updateSelection(user)
#         >>> print_overlays()
#         [+][-] A book

#     It works with persons and groups as well

#         >>> request.form['persons'] = ['joe']
#         >>> request.form['groups'] = ['friends']
#         >>> request.form['resources'] = []
#         >>> view._updateSelection(user)
#         >>> print_overlays()
#         [+][-] Friends
#         [+][+] Joe

#     Note that for persons (and only for persons), both the calendar, and the
#     timetable are visible.

#         >>> request.form['persons'] = []
#         >>> request.form['groups'] = []
#         >>> request.form['resources'] = []
#         >>> view._updateSelection(user)
#         >>> print_overlays()

#     You can choose the application calendar as well

#         >>> request.form['application'] = ''
#         >>> view._updateSelection(user)
#         >>> print_overlays()
#         [+][+] Application

#     Avoid duplicate additions:

#         >>> request.form['application'] = ''
#         >>> view._updateSelection(user)
#         >>> print_overlays()
#         [+][+] Application

#     And remove it

#         >>> del request.form['application']
#         >>> view._updateSelection(user)
#         >>> print_overlays()

#     """


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                                       optionflags=doctest.ELLIPSIS|
                                            doctest.REPORT_NDIFF|
                                            doctest.NORMALIZE_WHITESPACE))
    suite.addTest(doctest.DocTestSuite('schooltool.app.browser.overlay'))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2013 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Upgrade SchoolTool to generation 42.

Make schedule container and timetable timezones match app settings.
"""
import zope.lifecycleevent
from zope.app.generations.utility import getRootFolder, findObjectsProviding
from zope.annotation.interfaces import IAnnotations
from zope.component.hooks import getSite, setSite

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.app import getApplicationPreferences

SCHEDULES_KEY = 'schooltool.timetable.schedules'
TIMETABLES_KEY = 'schooltool.timetable.timetables'


def evolveScheduleContainers(app):
    annotations = IAnnotations(app)
    prefs = getApplicationPreferences(app)
    schedule_containers = app[SCHEDULES_KEY]
    for container in schedule_containers.values():
        if container.timezone != prefs.timezone:
            container.timezone = prefs.timezone
            zope.lifecycleevent.modified(container)


def evolveTimetableContainers(app):
    annotations = IAnnotations(app)
    prefs = getApplicationPreferences(app)
    schedule_containers = app[TIMETABLES_KEY]
    for container in schedule_containers.values():
        for timetable in container.values():
            if timetable.timezone != prefs.timezone:
                timetable.timezone = prefs.timezone
                zope.lifecycleevent.modified(timetable)


def evolve(context):
    root = getRootFolder(context)

    old_site = getSite()
    apps = list(findObjectsProviding(root, ISchoolToolApplication))
    for app in apps:
        setSite(app)
        evolveScheduleContainers(app)
        evolveTimetableContainers(app)

    setSite(old_site)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import

import sys

import celery.app.defaults
import celery.loaders.default
try:
    # Celery 2
    from celery.app.abstract import configurated as StartStopStep
    logger = None
except:
    # Celery 3
    from celery.bootsteps import StartStopStep
    from celery.utils.log import worker_logger as logger

import zope.configuration.config
import zope.configuration.xmlconfig
from zope.app.publication.zopepublication import ZopePublication
from zope.component import provideUtility
from ZODB.interfaces import IDatabase
from ZODB.ActivityMonitor import ActivityMonitor

from schooltool.app.main import SchoolToolMachinery, setLanguage
from schooltool.app.interfaces import ISchoolToolApplication


SCHOOLTOOL_CONFIG_NAMESPACES = {
    "SCHOOLTOOL": {
        "CONFIG": celery.app.defaults.Option('schooltool.conf', type="string"),
        "RETRY_DB_CONFLICTS": celery.app.defaults.Option(3, type="int"),
        }
    }
celery.app.defaults.NAMESPACES.update(SCHOOLTOOL_CONFIG_NAMESPACES)
celery.app.defaults.DEFAULTS.update(
    dict((key, value.default)
         for key, value in celery.app.defaults.flatten(SCHOOLTOOL_CONFIG_NAMESPACES)))


ACTIVE_MACHINERY = None


class ConfiguratedSchoolToolMachinery(StartStopStep,
                                      SchoolToolMachinery):

    _configured = False
    app = None
    db = None
    options = None

    def __init__(self, celery_app, **kwargs):
        self.app = celery_app
        self.logger = logger or self.app.log.get_default_logger()

    def loadOptions(self):
        options = self.Options()
        options.config_file = self.app.conf.SCHOOLTOOL_CONFIG
        if not options.config_file:
            self.logger.error('No configuration file given')
            sys.exit(1)
        options.config, handler = self.readConfig(options.config_file)
        if options.config.database.config.storage is None:
            raise Exception('No ZODB storage configured')
        return options

    def openDB(self, options):
        if self.db is not None:
            return self.db
        db_configuration = options.config.database
        db = db_configuration.open()
        connection = db.open()
        root = connection.root()
        app = root.get(ZopePublication.root_name)
        if app is None or not ISchoolToolApplication.providedBy(app):
            connection.close()
            return None
        connection.close()
        provideUtility(db, IDatabase)
        db.setActivityMonitor(ActivityMonitor())
        self.db = db
        return self.db

    def start(self):
        if self.db is not None:
            return # already running
        if not self._configured:
            self.logger.info('SchoolTool: Loading configuration')
            self.options = self.loadOptions()
            self.configure(self.options)
            self._configured = True
        self.db = self.openDB(self.options)

    def stop(self):
        if self.db is not None:
            self.db.close()
            self.db = None


class SchoolToolReportMachinery(ConfiguratedSchoolToolMachinery):

    def configure(self, options):
        self.options = options
        self.configureComponents(
            options,
            site_zcml=options.config.report_server_definition)
        setLanguage(options.config.lang)
        self.configureReportlab(options.config.reportlab_fontdir)


class GlobalMachinery(object):

    def init_machinery(self):
        global ACTIVE_MACHINERY
        if ACTIVE_MACHINERY:
            ACTIVE_MACHINERY.stop()
        ACTIVE_MACHINERY = self.machinery_factory(self.app)

    @property
    def machinery(self):
        global ACTIVE_MACHINERY
        if not ACTIVE_MACHINERY:
            self.init_machinery()
        return ACTIVE_MACHINERY


class ZopeLoader(celery.loaders.default.Loader,
                 GlobalMachinery):

    machinery_factory = ConfiguratedSchoolToolMachinery

    def on_worker_init(self, *args, **kw):
        super(ZopeLoader, self).on_worker_init(*args, **kw)

    def on_worker_process_init(self, *args, **kw):
        super(ZopeLoader, self).on_worker_process_init(*args, **kw)
        self.init_machinery()
        self.machinery.start()

    def on_task_init(self, *args, **kw):
        super(ZopeLoader, self).on_task_init(*args, **kw)

    def on_process_cleanup(self, *args, **kw):
        super(ZopeLoader, self).on_process_cleanup(*args, **kw)

    def on_worker_shutdown(self):
        super(ZopeLoader, self).on_worker_shutdown()
        self.machinery.stop()


class ReportLoader(ZopeLoader):
    machinery_factory = SchoolToolReportMachinery


def open_schooltool_db():
    global ACTIVE_MACHINERY
    if ACTIVE_MACHINERY is None:
        return None
    ACTIVE_MACHINERY.start()
    return ACTIVE_MACHINERY.db


def close_schooltool_db():
    global ACTIVE_MACHINERY
    if ACTIVE_MACHINERY is None:
        return
    ACTIVE_MACHINERY.stop()



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Basic person browser views.
"""

from zope.interface import Interface, implements
from zope.container.interfaces import INameChooser
from zope.app.form.browser.interfaces import ITerms
from zope.authentication.interfaces import IUnauthenticatedPrincipal
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.cachedescriptors.property import Lazy
from zope.component import adapts
from zope.component import getUtility, queryMultiAdapter, getMultiAdapter
from zope.i18n import translate
from z3c.form import form, field, button, validator
from z3c.form.interfaces import DISPLAY_MODE
from zope.interface import invariant, Invalid
from zope.interface import directlyProvides
from zope.intid.interfaces import IIntIds
from zope.schema import Password, TextLine, Choice, List
from zope.schema import ValidationError
from zope.schema.interfaces import ITitledTokenizedTerm
from zope.traversing.browser.absoluteurl import absoluteURL
from zope.security.checker import canAccess
from zope.proxy import sameProxiedObjects

from reportlab.lib import units
import z3c.form.interfaces
import zc.table.column
from zc.table.interfaces import ISortableColumn
from z3c.form.validator import SimpleFieldValidator

from schooltool.app.browser.app import ActiveSchoolYearContentMixin
from schooltool.app.browser.app import RelationshipViewBase
from schooltool.app.browser.app import EditRelationships
from schooltool.app.browser.app import RelationshipAddTableMixin
from schooltool.app.browser.app import RelationshipRemoveTableMixin
from schooltool.app.browser.report import ReportPDFView
from schooltool.app.browser.report import DefaultPageTemplate
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.interfaces import IApplicationPreferences
from schooltool.common.inlinept import InlineViewPageTemplate
from schooltool.common.inlinept import InheritTemplate
from schooltool.basicperson.interfaces import IDemographics
from schooltool.basicperson.interfaces import IDemographicsFields
from schooltool.basicperson.interfaces import IBasicPerson
from schooltool.contact.interfaces import IContactable
from schooltool.group.interfaces import IGroupContainer
from schooltool.person.interfaces import IPerson, IPersonFactory
from schooltool.person.browser.person import PersonTable, PersonTableFormatter
from schooltool.person.browser.person import PersonTableFilter
from schooltool.schoolyear.interfaces import ISchoolYearContainer
from schooltool.report.report import OldReportTask
from schooltool.report.browser.report import RequestRemoteReportDialog
from schooltool.skin.containers import TableContainerView
from schooltool.skin import flourish
from schooltool.skin.flourish.interfaces import IViewletManager
from schooltool.skin.flourish.form import FormViewlet
from schooltool.skin.flourish.viewlet import Viewlet, ViewletManager
from schooltool.skin.flourish.content import ContentProvider
from schooltool import table
from schooltool.table.catalog import IndexedLocaleAwareGetterColumn
from schooltool.task.interfaces import IMessageContainer
from schooltool.task.tasks import MessageCatalog
from schooltool.task.tasks import getLastMessagesReadTime
from schooltool.task.tasks import markMessagesRead
from schooltool.task.browser.task import MessageColumn
from schooltool.term.interfaces import IDateManager

from schooltool.common import SchoolToolMessage as _



class BasicPersonContainerView(TableContainerView):
    """A Person Container view."""
    template = ViewPageTemplateFile("templates/container.pt")
    delete_template = ViewPageTemplateFile("templates/person_container_delete.pt")

    index_title = _("Person index")

    def isDeletingHimself(self):
        person = IPerson(self.request.principal, None)
        return person in self.itemsToDelete

    @property
    def schoolyears(self):
        app = ISchoolToolApplication(None)
        syc = ISchoolYearContainer(app)
        return syc


class FlourishBasicPersonContainerView(flourish.page.Page):
    """A Person Container view."""

    content_template = InlineViewPageTemplate('''
      <div tal:content="structure context/schooltool:content/ajax/table" />
    ''')


class PersonContainerLinks(flourish.page.RefineLinksViewlet):
    """Person container links viewlet."""


class PersonLinks(flourish.page.RefineLinksViewlet):
    """Person links viewlet."""

    @property
    def title(self):
        return self.context.title


class PersonImportLinks(flourish.page.RefineLinksViewlet):
    """Person container import links viewlet."""


class PersonSettingsLinks(flourish.page.RefineLinksViewlet):
    """Person settings links viewlet."""


class PersonActionsLinks(flourish.page.RefineLinksViewlet):
    """Person actions links viewlet."""


class IPersonAddForm(IBasicPerson):

    group = Choice(
        title=_(u"Group"),
        description=_(u"You can select one group membership now.  Manage multiple memberships after account creation."),
        source="schooltool.basicperson.group_vocabulary",
        required=False)

    advisor = Choice(
        title=_(u"Advisor"),
        source="schooltool.basicperson.advisor_vocabulary",
        required=False)

    username = TextLine(
        title=_("Username"),
        description=_(u"Cannot begin with '+' or '@,' contain non-ascii characters or '/.'"),
        required=True)

    password = Password(
        title=_("Password"),
        description=_(u"Users cannot log in until a password is assigned."),
        required=False)

    confirm = Password(
        title=_("Confirm password"),
        required=False)

    @invariant
    def isPasswordConfirmed(person):
        if person.password != person.confirm:
            raise Invalid(_(u"Passwords do not match"))


class IPhotoField(Interface):

    photo = flourish.fields.Image(
        title=_('Photo'),
        description=_('An image file that will be converted to a JPEG no larger than 99x132 pixels (3:4 aspect ratio). Uploaded images must be JPEG or PNG files smaller than 10 MB'),
        size=(99,132),
        format='JPEG',
        required=False)


class PersonAddFormAdapter(object):
    implements(IPersonAddForm)
    adapts(IBasicPerson)

    def __init__(self, context):
        self.__dict__['context'] = context

    password = None

    def __setattr__(self, name, value):
        if name == 'password':
            self.context.setPassword(value)
        else:
            setattr(self.context, name, value)

    def __getattr__(self, name):
        return getattr(self.context, name)


class PhotoFieldFormAdapter(object):

    adapts(IBasicPerson)
    implements(IPhotoField)

    def __init__(self, context):
        self.__dict__['context'] = context

    def __setattr__(self, name, value):
        setattr(self.context, name, value)

    def __getattr__(self, name):
        return getattr(self.context, name)


class UsernameAlreadyUsed(ValidationError):
    __doc__ = _("This username is already in use")


class UsernameBadName(ValidationError):
    __doc__ = _("Names cannot begin with '+' or '@' or contain '/'")


class UsernameNonASCII(ValidationError):
    __doc__ = _("Usernames cannot contain non-ascii characters")


class UsernameValidator(SimpleFieldValidator):

    def validate(self, username):
        super(UsernameValidator, self).validate(username)
        if username is not None:
            if username in self.context:
                raise UsernameAlreadyUsed(username)
            try:
                INameChooser(self.context).checkName(username, None)
            except ValueError:
                raise UsernameBadName(username)
        # XXX: this has to be fixed
        # XXX: SchoolTool should handle UTF-8
        try:
            username.encode('ascii')
        except UnicodeEncodeError:
            raise UsernameNonASCII(username)


validator.WidgetValidatorDiscriminators(UsernameValidator,
                                        field=IPersonAddForm['username'])


class PasswordsDontMatch(ValidationError):
    __doc__ = _(u"Passwords do not match")


class PasswordValidator(SimpleFieldValidator):

    def validate(self, value):
        # XXX: hack to display the password error next to the widget!
        super(PasswordValidator, self).validate(value)
        confirm_widget = self.view.widgets['confirm']
        confirm_value = self.request.get(confirm_widget.name)
        if value is not None and value != confirm_value:
            raise PasswordsDontMatch()


validator.WidgetValidatorDiscriminators(PasswordValidator,
                                        field=IPersonAddForm['password'])


class PersonForm(object):

    formErrorsMessage = _('Please correct the marked fields below.')

    def generateExtraFields(self):
        field_descriptions = IDemographicsFields(ISchoolToolApplication(None))
        fields = field.Fields()
        if IBasicPerson.providedBy(self.context):
            limit_keys = [group.__name__ for group in self.context.groups]
        else:
            limit_keys = []
        for field_desc in field_descriptions.filter_keys(limit_keys):
            fields += field_desc.makeField()
        return fields


class PersonView(PersonForm, form.DisplayForm):

    template = ViewPageTemplateFile('templates/person_view.pt')

    @property
    def label(self):
        return self.context.title

    def update(self):
        self.fields = field.Fields(IBasicPerson)
        self.fields += self.generateExtraFields()
        super(PersonView, self).update()

    def __call__(self):
        self.update()
        return self.render()


class FlourishPersonView(flourish.page.Page):
    """Person index.html view."""


class FlourishPersonInfo(flourish.page.Content):
    body_template = ViewPageTemplateFile('templates/f_person_view_details.pt')

    def markRead(self):
        auth_person = IPerson(self.request.principal, None)
        if (auth_person is None or
            not sameProxiedObjects(auth_person, self.context)):
            return # not looking at his own homepage
        markMessagesRead(self.context)

    def update(self):
        self.markRead()
        super(FlourishPersonInfo, self).update()

    @property
    def canModify(self):
        return canAccess(self.context.__parent__, '__delitem__')

    @property
    def done_link(self):
        done_link = self.request.get('done_link', None)
        if done_link is not None:
            return done_link
        app = ISchoolToolApplication(None)
        return absoluteURL(app, self.request) + '/persons'


class PersonAddFormBase(PersonForm, form.AddForm):
    """Person add form for basic persons."""

    def update(self):
        self.fields = field.Fields(IPersonAddForm)
        self.fields += self.generateExtraFields()
        super(PersonAddFormBase, self).update()

    def updateActions(self):
        super(PersonAddFormBase, self).updateActions()
        self.actions['add'].addClass('button-ok')

    @button.buttonAndHandler(_('Add'), name='add')
    def handleAdd(self, action):
        super(PersonAddFormBase, self).handleAdd.func(self, action)

    def create(self, data):
        person = self._factory(data['username'], data['first_name'],
                               data['last_name'])
        data.pop('confirm')
        group = data.pop('group')
        advisor = data.pop('advisor')
        form.applyChanges(self, person, data)
        if group is not None:
            person.groups.add(group)
        if advisor is not None:
            person.advisors.add(advisor)
        self._person = person
        return person

    def nextURL(self):
        return absoluteURL(self._person, self.request)

    @property
    def _factory(self):
        return getUtility(IPersonFactory)

    def add(self, person):
        """Add `person` to the container.

        Uses the username of `person` as the object ID (__name__).
        """
        name = person.username
        self.context[name] = person
        return person


class PersonAddView(PersonAddFormBase):
    """Person add form for basic persons."""
    label = _("Add new person")
    template = ViewPageTemplateFile('templates/person_add.pt')

    def updateActions(self):
        super(PersonAddView, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @button.buttonAndHandler(_('Add'), name='add')
    def handleAdd(self, action):
        super(PersonAddView, self).handleAdd.func(self, action)

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)


class ITempPersonList(Interface):
    usernames = List(
        title=_('Usernames'),
        value_type=TextLine(
            title=_('Username'),
            required=False),
        required=False,
        )


class TempPersonList(object):
    implements(ITempPersonList)

    def __init__(self):
        self.usernames = []


class MultiplePersonAddView(form.Form):
    template = ViewPageTemplateFile('templates/person_add_multiple.pt')
    label = _("Add multiple persons")
    fields = field.Fields(ITempPersonList)
    addform = None
    temp_content = None

    def __init__(self, *args, **kw):
        self.temp_content = TempPersonList()
        super(MultiplePersonAddView, self).__init__(*args, **kw)

    def getContent(self):
        return self.temp_content

    def buildAddForm(self):
        self.addform = PersonAddSubForm(self.context, self.request, self)
        n = len(self.widgets['usernames'].value)
        self.addform.prefix = 'addform_%d.' % n

    def updateWidgets(self):
        super(MultiplePersonAddView, self).updateWidgets()
        self.widgets['usernames'].mode = z3c.form.interfaces.HIDDEN_MODE
        self.buildAddForm()

    def update(self):
        super(MultiplePersonAddView, self).update()
        data, errors = self.extractData(setErrors=False)
        form.applyChanges(self, self.getContent(), data)
        self.addform.update()

    def appendAdded(self, username):
        usernames = self.widgets['usernames'].value
        usernames.append(username)
        self.widgets['usernames'].value = usernames
        content = self.getContent()
        content.usernames = usernames

    def render(self):
        if self.addform._finishedAdd:
            self.appendAdded(self.addform._person.__name__)
            # And build a fresh add form.
            self.buildAddForm()
            self.addform.update()
        return super(MultiplePersonAddView, self).render()

    @property
    def addedPersons(self):
        app_persons = ISchoolToolApplication(None)['persons']
        content = self.getContent()
        if not content.usernames:
            return []
        return [app_persons[username]
                for username in content.usernames
                if username in app_persons]

    def updateActions(self):
        super(MultiplePersonAddView, self).updateActions()
        self.actions['done'].addClass('button-neutral')

    @button.buttonAndHandler(_("Done"))
    def handle_done_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)


class FlourishMultiplePersonAddView(MultiplePersonAddView):
    template = InheritTemplate(flourish.page.Page.template)

    def buildAddForm(self):
        self.addform = FlourishPersonAddSubForm(
            self.context, self.request, self)
        n = len(self.widgets['usernames'].value)
        self.addform.prefix = 'addform_%d.' % n


class PersonAddSubForm(PersonAddFormBase):
    template = ViewPageTemplateFile('templates/person_add_subform.pt')

    def __init__(self, context, request, parentForm):
        self.parentForm = parentForm
        super(PersonAddSubForm, self).__init__(context, request)

    def update(self):
        super(PersonAddSubForm, self).update()
        for action in self.parentForm.actions.executedActions:
            handler = queryMultiAdapter(
                (self, self.request, self.getContent(), action),
                interface=z3c.form.interfaces.IActionHandler)
            if handler is not None:
                handler()


class FlourishPersonAddSubForm(PersonAddSubForm):
    template = ViewPageTemplateFile('templates/f_person_add_subform.pt')


class PersonEditView(PersonForm, form.EditForm):
    """Edit form for basic person."""
    form.extends(form.EditForm)
    template = ViewPageTemplateFile('templates/person_add.pt')

    def update(self):
        self.fields = field.Fields(IBasicPerson)
        self.fields += self.generateExtraFields()
        super(PersonEditView, self).update()

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    def updateActions(self):
        super(PersonEditView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @property
    def label(self):
        return _(u'Change information for ${fullname}',
                 mapping={'fullname': self.context.title})


class FlourishPersonEditView(flourish.page.Page, PersonEditView):

    label = None

    def update(self):
        self.buildFieldsetGroups()
        self.fields = field.Fields(IBasicPerson)
        self.fields += field.Fields(IPhotoField)
        self.fields += self.generateExtraFields()
        form.EditForm.update(self)

    @button.buttonAndHandler(_('Apply'), name='apply')
    def handleApply(self, action):
        super(FlourishPersonEditView, self).handleApply.func(self, action)
        # XXX: hacky sucessful submit check
        if (self.status == self.successMessage or
            self.status == self.noChangesMessage):
            self.request.response.redirect(self.nextURL())

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        self.request.response.redirect(self.nextURL())

    def nextURL(self):
        url = absoluteURL(self.context, self.request)
        return url

    def makeRows(self, fields, cols=1):
        rows = []
        while fields:
            rows.append(fields[:cols])
            fields = fields[cols:]
        return rows

    def makeFieldSet(self, fieldset_id, legend, fields, cols=1):
        result = {
            'id': fieldset_id,
            'legend': legend,
            }
        result['rows'] = self.makeRows(fields, cols)
        return result

    def buildFieldsetGroups(self):
        self.fieldset_groups = {
            'full_name': (
                _('Full Name'),
                ['prefix', 'first_name', 'middle_name', 'last_name',
                 'suffix', 'preferred_name']),
            'details': (
                _('Details'), ['gender', 'birth_date', 'photo']),
            'demographics': (
                _('Demographics'), list(self.generateExtraFields())),
            }
        self.fieldset_order = (
            'full_name', 'details', 'demographics')

    def fieldsets(self):
        result = []
        for fieldset_id in self.fieldset_order:
            legend, fields = self.fieldset_groups[fieldset_id]
            result.append(self.makeFieldSet(
                    fieldset_id, legend, list(fields)))
        return result


class PersonTerm(object):
    """A term for displaying a person."""
    implements(ITitledTokenizedTerm)

    def __init__(self, value):
        self.title = value.title
        self.token = value.__name__
        self.value = value


class TermsBase(object):
    """Base terms implementation."""

    implements(ITerms)

    def factory(self, value):
        raise NotImplementedError(
            "Term Factory must be provided by inheriting classes.")

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def getTerm(self, value):
        if value not in self.context:
            raise LookupError(value)
        return self.factory(value)

    def getValue(self, token):
        if token not in self.context:
            raise LookupError(token)
        return token


class PersonTerms(TermsBase):
    """Displaying persons."""
    factory = PersonTerm

    def getValue(self, token):
        if token not in self.context:
            raise LookupError(token)
        app = ISchoolToolApplication(None)
        return app['persons'][token]


class GroupTerm(object):
    """A term for displaying a group."""
    implements(ITitledTokenizedTerm)

    def __init__(self, value):
        self.title = value.title
        self.token = value.__name__
        self.value = value


class GroupTerms(TermsBase):
    """Displaying groups."""

    factory = GroupTerm


class PersonAdvisorView(RelationshipViewBase):
    """View class for adding/removing advisors to/from a person."""

    __used_for__ = IBasicPerson

    current_title = _("Current Advisors")
    available_title = _("Add Advisors")

    @property
    def title(self):
        return _("Advisors of ${person}",
            mapping={'person': self.context.title})

    def getSelectedItems(self):
        """Return a list of current advisors."""
        return self.context.advisors

    def getAvailableItemsContainer(self):
        return ISchoolToolApplication(None)['persons']

    def getCollection(self):
        return self.context.advisors


class EditPersonRelationships(EditRelationships):

    def getAvailableItemsContainer(self):
        return ISchoolToolApplication(None)['persons']


class FlourishPersonAdvisorView(EditPersonRelationships):

    current_title = _('Current advisors')
    available_title = _('Available advisors')

    def getSelectedItems(self):
        return self.context.advisors

    def getCollection(self):
        return self.context.advisors


class PersonAdviseeView(RelationshipViewBase):
    """View class for adding/removing advisees to/from a person."""

    __used_for__ = IBasicPerson

    current_title = _("Current Advisees")
    available_title = _("Add Advisees")

    @property
    def title(self):
        return _("Advisees of ${person}",
            mapping={'person': self.context.title})

    def getSelectedItems(self):
        """Return a list of current advisees."""
        return self.context.advisees

    def getAvailableItemsContainer(self):
        return ISchoolToolApplication(None)['persons']

    def getCollection(self):
        return self.context.advisees


class FlourishPersonAdviseeView(EditPersonRelationships):

    current_title = _("Current advisees")
    available_title = _("Available advisees")

    def getSelectedItems(self):
        return self.context.advisees

    def getCollection(self):
        return self.context.advisees


class IFlourishPersonInfoManager(IViewletManager):
    pass


class FlourishPersonInfoManager(ViewletManager):
    pass


class FlourishAdvisoryViewlet(Viewlet):
    """A viewlet showing the advisors/advisees of a person."""

    template = ViewPageTemplateFile('templates/f_advisoryViewlet.pt')

    @property
    def canModify(self):
        return canAccess(self.context.__parent__, '__delitem__')


class FlourishGeneralViewlet(FormViewlet):
    """A viewlet showing the core attributes of a person."""

    template = ViewPageTemplateFile('templates/f_generalViewlet.pt')
    mode = DISPLAY_MODE

    def getFields(self):
        field_descriptions = IDemographicsFields(ISchoolToolApplication(None))
        fields = field.Fields()
        for field_desc in field_descriptions.values():
            fields += field_desc.makeField()
        return fields

    @property
    def filtered_widgets(self):
        result = [widget for widget in self.widgets.values()
                  if widget.value]
        return result

    @property
    def heading(self):
        attrs = ['prefix', 'first_name', 'middle_name', 'last_name', 'suffix']
        values = []
        for attr in attrs:
            value = getattr(self.context, attr)
            if value:
                values.append(value)
        return ' '.join(values)

    def makeRow(self, attr, value):
        if value is None:
            value = u''
        return {
            'label': attr,
            'value': unicode(value),
            }

    @property
    def table(self):
        rows = []
        fields = field.Fields(IBasicPerson)
        fields += field.Fields(IPerson).select('username')
        for attr in fields:
            value = getattr(self.context, attr)
            if value:
                label = fields[attr].field.title
                if attr == 'gender':
                    vocabulary = IBasicPerson[attr].vocabulary
                    message = vocabulary.getTermByToken(value).title
                    value = translate(message, context=self.request)
                rows.append(self.makeRow(label, value))
        return rows

    @property
    def canModify(self):
        return canAccess(self.context.__parent__, '__delitem__')

    def update(self):
        self.fields = self.getFields()
        super(FlourishGeneralViewlet, self).update()

    def has_photo(self):
        return self.context.photo is not None

    def table_class(self):
        result = 'person-view-demographics'
        if self.has_photo():
            result += ' show-photo'
        return result


###############  Base class of all group-aware add views ################
class PersonAddViewBase(PersonAddFormBase):

    id = 'person-form'
    template = ViewPageTemplateFile('templates/person_form.pt')

    def makeRows(self, fields, cols=1):
        rows = []
        while fields:
            rows.append(fields[:cols])
            fields = fields[cols:]
        return rows

    def makeFieldSet(self, fieldset_id, legend, fields, cols=1):
        result = {
            'id': fieldset_id,
            'legend': legend,
            }
        result['rows'] = self.makeRows(fields, cols)
        return result

    def fieldsets(self):
        result = []
        sources = [
            (self.base_id, self.base_legend, list(self.getBaseFields())),
            (self.demo_id, self.demo_legend, list(self.getDemoFields())),
            ]
        for fieldset_id, legend, fields in sources:
            result.append(self.makeFieldSet(fieldset_id, legend, fields, 2))
        return result

    def getDemoFields(self):
        fields = field.Fields()
        dfs = IDemographicsFields(ISchoolToolApplication(None))
        keys = self.group_id and [self.group_id] or []
        for field_desc in dfs.filter_keys(keys):
            fields += field_desc.makeField()
        return fields

    def getBaseFields(self):
        return field.Fields(IPersonAddForm).omit('group')

    def updateActions(self):
        super(PersonAddViewBase, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @button.buttonAndHandler(_('Add'))
    def handleAdd(self, action):
        super(PersonAddViewBase, self).handleAdd.func(self, action)

    @button.buttonAndHandler(_('Cancel'))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    def create(self, data):
        username = data.get('username')
        first_name = data.get('first_name')
        last_name = data.get('last_name')
        person = self._factory(username, first_name, last_name)
        data.pop('confirm')
        form.applyChanges(self, person, data)
        group = None
        syc = ISchoolYearContainer(ISchoolToolApplication(None))
        active_schoolyear = syc.getActiveSchoolYear()
        if active_schoolyear is not None:
            if self.group_id:
                group = IGroupContainer(active_schoolyear).get(self.group_id)
            else:
                group = data.get('group')
        if group is not None:
            person.groups.add(group)
        advisor = data.get('advisor')
        if advisor is not None:
            person.advisors.add(advisor)
        self._person = person
        return person

    def update(self):
        self.fields = self.getBaseFields()
        self.fields += self.getDemoFields()
        self.updateWidgets()
        self.updateActions()
        self.actions.execute()

    def updateWidgets(self):
        super(PersonAddViewBase, self).updateWidgets()


class FlourishPersonAddView(PersonAddViewBase):
    template = InheritTemplate(flourish.page.Page.template)
    page_template = InheritTemplate(flourish.page.Page.page_template)

    fieldset_groups = None
    fieldset_order = None

    group_id = None

    def update(self):
        self.buildFieldsetGroups()
        PersonAddViewBase.update(self)
        self.widgets['birth_date'].addClass('birth-date-field')

    def getBaseFields(self):
        result = field.Fields(IPersonAddForm)
        if self.group_id:
            result = result.omit('group')
        result += field.Fields(IPhotoField)
        return result

    def buildFieldsetGroups(self):
        relationship_fields = ['advisor']
        if not self.group_id:
            relationship_fields[0:0] = ['group']
        self.fieldset_groups = {
            'full_name': (
                _('Full Name'),
                ['prefix', 'first_name', 'middle_name', 'last_name',
                 'suffix', 'preferred_name']),
            'user': (
                _('User'), ['username', 'password', 'confirm']),
            'details': (
                _('Details'), ['gender', 'birth_date', 'photo']),
            'demographics': (
                _('Demographics'), list(self.getDemoFields())),
            'relationships': (
                _('Relationships'), relationship_fields),
            }
        self.fieldset_order = (
            'full_name', 'details', 'demographics',
            'relationships', 'user')

    def fieldsets(self):
        result = []
        for fieldset_id in self.fieldset_order:
            legend, fields = self.fieldset_groups[fieldset_id]
            result.append(self.makeFieldSet(
                    fieldset_id, legend, list(fields)))
        return result

    @button.buttonAndHandler(_('Submit'), name='add')
    def handleSubmit(self, action):
        super(FlourishPersonAddView, self).handleAdd.func(self, action)

    @button.buttonAndHandler(_('Submit and add'), name='submitadd')
    def handleSubmitAndAdd(self, action):
        super(FlourishPersonAddView, self).handleAdd.func(self, action)
        if self._finishedAdd:
            self.request.response.redirect(self.action)

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    def updateActions(self):
        super(FlourishPersonAddView, self).updateActions()
        self.actions['submitadd'].addClass('button-neutral')


###############  Group-aware add views ################
class TeacherAddView(PersonAddViewBase):

    group_id = 'teachers'
    base_id = 'base-data'
    base_legend = _('Teacher identification')
    demo_id = 'demo-data'
    demo_legend = _('Teacher demographics')
    label = _('Add new teacher')


class StudentAddView(PersonAddViewBase):

    group_id = 'students'
    base_id = 'base-data'
    base_legend = _('Student identification')
    demo_id = 'demo-data'
    demo_legend = _('Student demographics')
    label = _('Add new student')


class AdministratorAddView(PersonAddViewBase):

    group_id = 'administrators'
    base_id = 'base-data'
    base_legend = _('Administrator identification')
    demo_id = 'demo-data'
    demo_legend = _('Administrator demographics')
    label = _('Add new administrator')


class AddPersonViewlet(object):

    def hasSchoolYear(self):
        app = ISchoolToolApplication(None)
        syc = ISchoolYearContainer(app)
        sy = syc.getActiveSchoolYear()
        return sy is not None


class PersonTitle(ContentProvider):
    render = InlineViewPageTemplate('''
        <span tal:replace="view/title"></span>
    '''.strip())

    def title(self):
        person = self.context
        return person.title


class BasicPersonTable(PersonTable):

    def __init__(self, *args, **kw):
        PersonTable.__init__(self, *args, **kw)
        self.css_classes = {'table': ' data persons-table'}

    def columns(self):
        cols = getUtility(IPersonFactory).columns()
        username = IndexedLocaleAwareGetterColumn(
            index='__name__',
            name='username',
            title=_(u'Username'),
            getter=lambda i, f: i.__name__,
            subsort=True)
        return cols + [username]


class PersonListTable(BasicPersonTable):

    @property
    def source(self):
        return ISchoolToolApplication(None)['persons']

    def items(self):
        return self.indexItems(self.context)


class BasicPersonAddRelationshipTable(RelationshipAddTableMixin,
                                      BasicPersonTable):
    pass


class BasicPersonRemoveRelationshipTable(RelationshipRemoveTableMixin,
                                         BasicPersonTable):
    pass


class BasicPersonTableFormatter(PersonTableFormatter):

    def columns(self):
        cols = list(reversed(PersonTableFormatter.columns(self)))
        username = IndexedLocaleAwareGetterColumn(
            index='__name__',
            name='username',
            title=_(u'Username'),
            getter=lambda i, f: i.__name__,
            subsort=True)
        return cols + [username]

    def makeFormatter(self):
        formatter = PersonTableFormatter.makeFormatter(self)
        formatter.cssClasses['table'] = 'data persons-table'
        return formatter


class FlourishManagePeopleOverview(flourish.page.Content,
                                   ActiveSchoolYearContentMixin):

    body_template = ViewPageTemplateFile(
        'templates/f_manage_people_overview.pt')

    built_in_groups = ('administrators', 'clerks', 'manager', 'teachers',
                       'students')

    @property
    def groups(self):
        return IGroupContainer(self.schoolyear, None)

    @property
    def persons(self):
        return self.context['persons']

    @property
    def school_name(self):
        preferences = IApplicationPreferences(self.context)
        return preferences.title

    def persons_url(self):
        return self.url_with_schoolyear_id(self.context, view_name='persons')


class FlourishRequestPersonIDCardView(RequestRemoteReportDialog):

    report_task = OldReportTask
    report_builder = 'person_id_card.pdf'


class IDCardsPageTemplate(DefaultPageTemplate):

    template = ViewPageTemplateFile('templates/f_id_cards_template.pt',
                                    content_type="text/xml")


class FlourishPersonIDCardsViewBase(ReportPDFView):

    template=ViewPageTemplateFile('templates/f_person_id_cards.pt')
    title = _('ID Cards')
    COLUMNS = 2
    ROWS = 4
    # All of the following are cm
    CARD = {
        'height': 5.4,
        'width': 8.57,
        }
    TITLE = {
        'height': 1.1,
        'padding': 0.06, # from the edge of the title frame
        }
    DEMOGRAPHICS = {
        'height': 3.2,
        'margin-left': 0.55, # from the edge of the card
        'margin-bottom': 0.75, # from the edge of the card
        'width': 5.05,
        }
    PHOTO = {
        'height': 3.2,
        'width': 2.4,
        }
    FOOTER = {
        'height': 0.5,
        }
    LEFT_BASE = 1.7
    TOP_BASE = 7.3
    COLUMN_WIDTH = 9.6
    ROW_HEIGHT = 6.1

    def __init__(self, *args, **kw):
        super(FlourishPersonIDCardsViewBase, self).__init__(*args, **kw)
        app = ISchoolToolApplication(None)
        preferences = IApplicationPreferences(app)
        self.schoolName = preferences.title

    @property
    def top(self):
        page_height = self.pageSize[1]
        return (page_height/units.cm) - self.TOP_BASE

    @property
    def left(self):
        return self.LEFT_BASE

    def persons(self):
        """Returns a list of getPersonData calls"""
        raise NotImplementedError('do this in subclass')

    def getPersonData(self, person):
        demographics = IDemographics(person)
        contact_title = None
        contact_phone = None
        contacts = list(IContactable(person).contacts)
        if contacts:
            contact = contacts[0]
            contact_title = contact.title
            phones = [
                contact.home_phone,
                contact.work_phone,
                contact.mobile_phone,
                ]
            if phones:
                contact_phone = phones[0]
        return {
            'title': person.title,
            'username': person.username,
            'ID': demographics.get('ID'),
            'birth_date': person.birth_date,
            'contact_title': contact_title,
            'contact_phone': contact_phone,
            'photo': person.photo,
            }

    def titleVerticalAlign(self, person):
        result = 0
        if len(person['title']) < 44:
            result = 0.25
        return '%fcm' % result

    @Lazy
    def total_cards_in_page(self):
        return self.COLUMNS * self.ROWS

    def insertBreak(self, repeat):
        return repeat['person'].number() % self.total_cards_in_page  == 0

    def personFrame(self, repeat):
        return repeat['person'].index() % self.total_cards_in_page

    def cards(self):
        result = []
        for card_index in range(self.total_cards_in_page):
            card_column = card_index % self.COLUMNS
            card_row = card_index / self.COLUMNS
            card_x1 = self.left + (self.COLUMN_WIDTH * card_column)
            card_y1 = self.top - (self.ROW_HEIGHT * card_row)
            card_info = {
                'index': card_index,
                'x1': card_x1,
                'y1': card_y1,
                'height': self.CARD['height'],
                'width': self.CARD['width'],
                }
            result.append(card_info)
        return result

    @Lazy
    def cardFrames(self):
        result = {}
        for card_info in self.cards():
            info = {
                'outter': self.getOutterFrame(card_info),
                'title': self.getTitleFrame(card_info),
                'demographics': self.getDemographicsFrame(card_info),
                'footer': self.getFooterFrame(card_info),
                }
            result[card_info['index']] = info
        return result

    def getOutterFrame(self, card_info):
        frame_info = card_info.copy()
        frame_info['id'] = 'outter_%d' % card_info['index']
        frame_info['maxWidth'] = card_info['width']
        frame_info['maxHeight'] = card_info['height']
        return frame_info

    def getTitleFrame(self, card_info):
        y1 = card_info['y1'] + (card_info['height'] - self.TITLE['height'])
        return {
            'id': 'title_%d' % card_info['index'],
            'x1': card_info['x1'],
            'y1': y1,
            'width': card_info['width'],
            'maxWidth': card_info['width'] - self.TITLE['padding'],
            'height': self.TITLE['height'],
            'maxHeight': self.TITLE['height'] - self.TITLE['padding'],
            }

    def getDemographicsFrame(self, card_info):
        demo_info = self.DEMOGRAPHICS
        x1 = card_info['x1'] + demo_info['margin-left']
        y1 = card_info['y1'] +  demo_info['margin-bottom']
        return {
            'id': 'demographics_%d' % card_info['index'],
            'x1': x1,
            'y1': y1,
            'width': demo_info['width'],
            'maxWidth': demo_info['width'],
            'height': demo_info['height'],
            'maxHeight': demo_info['height'],
            }

    def getFooterFrame(self, card_info):
        x1 = card_info['x1']
        y1 = card_info['y1']
        return {
            'id': 'footer_%d' % card_info['index'],
            'x1': x1,
            'y1': y1,
            'width': card_info['width'],
            'maxWidth': card_info['width'],
            'height': self.FOOTER['height'],
            'maxHeight': self.FOOTER['height'],
            }


class FlourishPersonIDCardView(FlourishPersonIDCardsViewBase):

    @property
    def filename(self):
        return 'id_cards_%s_%s.pdf' % (
            self.context.last_name,  self.context.first_name)

    @property
    def title(self):
        return _('ID Card: ${person}',
                 mapping={'person': self.context.title})

    def persons(self):
        return [self.getPersonData(self.context)] * self.total_cards_in_page


def getUserViewlet(context, request, view, manager, name):
    principal = request.principal
    user = IPerson(principal, None)
    if user is None:
        return None
    viewlet = flourish.viewlet.lookupViewlet(
        user, request, view, manager, name=name)
    return viewlet


def getPersonActiveViewlet(person, request, view, manager):
    user = IPerson(request.principal, None)
    if (user is not None and
        user.__name__ == person.__name__):
        return u"home"
    return flourish.page.getParentActiveViewletName(
        person, request, view, manager)


class PhotoView(flourish.widgets.ImageView):
    attribute = "photo"


class PersonContainerViewTableFilter(PersonTableFilter, ActiveSchoolYearContentMixin):

    template = ViewPageTemplateFile('templates/f_container_table_filter.pt')

    @property
    def done_link(self):
        app = ISchoolToolApplication(None)
        return self.url_with_schoolyear_id(app, 'manage')


class PersonProfilePDF(flourish.report.PlainPDFPage):

    name = _("Profile")
    message_title = _('profile report')

    def formatDate(self, date, format='mediumDate'):
        if date is None:
            return ''
        formatter = getMultiAdapter((date, self.request), name=format)
        return formatter()

    @property
    def scope(self):
        dtm = getUtility(IDateManager)
        today = dtm.today
        return self.formatDate(today)

    @property
    def subtitles_left(self):
        student_id = _('Username: ${username}',
                       mapping={'username': self.context.username})
        subtitles = [
            student_id,
            ]
        return subtitles

    @property
    def title(self):
        return self.context.title

    @property
    def base_filename(self):
        return '%s_%s_%s' % (
            self.context.last_name,  self.context.first_name, self.context.username)


class ProfileGeneralPart(flourish.report.PDFForm):

    title = _("General Information")

    def getFields(self):
        field_descriptions = IDemographicsFields(ISchoolToolApplication(None))
        fields = field.Fields()
        for field_desc in field_descriptions.values():
            fields += field_desc.makeField()
        return fields

    def update(self):
        self.fields = field.Fields(IBasicPerson)
        self.fields += field.Fields(IPhotoField)
        self.fields += field.Fields(IPerson).select('username')
        self.fields += self.getFields()
        super(ProfileGeneralPart, self).update()


class UserLinkViewlet(flourish.page.LinkViewlet):

    @property
    def url(self):
        user = IPerson(self.request.principal, None)
        link = self.link
        if not link or user is None:
            return None
        return "%s/%s" % (absoluteURL(user, self.request), self.link)


class PersonMessageFilter(table.table.DoNotFilter):

    @property
    def catalog(self):
        return self.manager.catalog

    @property
    def person(self):
        return IPerson(self.context)

    @property
    def active_person_intid(self):
        person = self.person
        int_ids = getUtility(IIntIds)
        person_intid = int_ids.queryId(person, None)
        return person_intid

    def filter(self, results):
        person_intid = self.active_person_intid
        if person_intid is None:
            return []
        recipient_index = self.catalog['recipient_ids']
        message_ids = recipient_index.apply({'any_of': set([person_intid])})
        results = [item for item in results
                   if item['id'] in message_ids]
        return results


class MessageTable(table.ajax.IndexedTable):

    @property
    def source(self):
        return IMessageContainer(ISchoolToolApplication(None))

    def columns(self):
        group = table.column.IndexedGetterColumn(
            name='group',
            title=_(u"Group"),
            getter=lambda i, f: i.group,
            cell_formatter=table.table.translate_cell_formatter,
            index='group',
            subsort=True)
        directlyProvides(group, ISortableColumn)
        updated = table.column.IndexedGetterColumn(
            name='updated_on',
            title=_("Received"),
            getter=lambda i, f: i.updated_on,
            cell_formatter=table.table.datetime_formatter,
            index='updated_on',
            subsort=True)
        directlyProvides(updated, ISortableColumn)
        message = MessageColumn(
            name='message',
            title=_("Message"))
        return [message, group, updated]

    def sortOn(self):
        return (('updated_on', True), )


class FlourishRequestPersonProfileView(RequestRemoteReportDialog):

    report_builder = 'person_profile.pdf'


class NewMessageIndicatorViewlet(flourish.page.LinkViewlet):

    css_class = 'new-messages'

    @property
    def authenticated_person(self):
        principal = getattr(self.request, 'principal', None)
        if principal is None:
            return None
        if IUnauthenticatedPrincipal.providedBy(principal):
            return None
        return IPerson(principal, None)

    @property
    def authenticated_person_intid(self):
        person = self.authenticated_person
        if person is None:
            return None
        int_ids = getUtility(IIntIds)
        person_intid = int_ids.queryId(person, None)
        return person_intid

    @property
    def new_messages(self):
        person = self.authenticated_person
        if person is None:
            return None
        person_intid = self.authenticated_person_intid
        if person_intid is None:
            return None
        dt = getLastMessagesReadTime(person)
        catalog = MessageCatalog.get()
        new_ids = catalog['updated_on'].apply(
            {'between': (dt, None)})
        my_ids = catalog['recipient_ids'].apply(
            {'any_of': set([person_intid])})
        new_messages = len(set(my_ids).intersection(new_ids))
        return new_messages

    @property
    def title(self):
        person = self.authenticated_person
        if person is None:
            return ''
        new_messages = self.new_messages
        if not new_messages:
            return ''
        title = _('(${number} new)',
                  mapping={'number': new_messages})
        return title

    @property
    def url(self):
        person = self.authenticated_person
        if person is None:
            return None
        url = absoluteURL(person, self.request) + '#messages'
        return url



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2007, 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""
SchoolTool run script.
"""

from __future__ import with_statement

import optparse
import os
import os.path

import paste.script.command


def parse_args():
    """Parse the command line arguments"""
    parser = optparse.OptionParser(usage="usage: %prog INSTANCE [options]")
    parser.add_option("--daemon",
                      action="store_true",
                      dest="start_daemon",
                      help="Run in daemon (background) mode")
    parser.add_option('--pid-file',
                      dest='pid_file',
                      metavar='FILENAME',
                      help="Save PID to file (if running in daemon mode)")
    parser.add_option('--log-file',
                      dest='log_file',
                      metavar='LOG_FILE',
                      help="Save output to the given log file (redirects stdout)")
    parser.add_option("--stop-daemon",
                      action="store_true",
                      dest="stop_daemon",
                      help="Stop a daemonized server")
    parser.add_option('--reload',
                      dest='reload',
                      action='store_true',
                      help="Use auto-restart file monitor")
    parser.add_option('--reload-interval',
                      dest='reload_interval',
                      default=1,
                      help="Seconds between checking files (low number can cause significant CPU usage)")
    parser.add_option('--monitor-restart',
                      dest='monitor_restart',
                      action='store_true',
                      help="Auto-restart server if it dies")
    parser.add_option("--status",
                      action="store_true",
                      dest="show_status",
                      help="Show the status of the (presumably daemonized) server")
    parser.add_option('--user',
                      dest='set_user',
                      metavar="USERNAME",
                      help="Set the user (usually only possible when run as root)")
    parser.add_option('--group',
                      dest='set_group',
                      metavar="GROUP",
                      help="Set the group (usually only possible when run as root)")
    options, args = parser.parse_args()

    if len(args) != 1:
        parser.error("""Missing instance to start up! You can create one using make-scooltool-instance.""")

    instance_root = os.path.abspath(args[0])
    conf_file = os.path.join(instance_root, "paste.ini")
    if not os.path.exists(conf_file):
        parser.error("This is not a schooltool instance: %s" % args[0])
    args[0] = conf_file

    if (options.start_daemon or
        options.stop_daemon or
        options.show_status):
        if not options.pid_file:
            options.pid_file = os.path.join(instance_root,
                                            "var", "schooltool.pid")
        if not options.log_file:
            options.log_file = os.path.join(instance_root,
                                            "log", "paster.log")
    return options, args


def set_default_celery_config():
    if 'CELERY_CONFIG_MODULE' not in os.environ:
        os.environ['CELERY_CONFIG_MODULE']='schooltool.task.config.zope'


def main():
    options, args = parse_args()
    conf_file = os.path.abspath(args[0])

    set_default_celery_config()

    extra_options = []
    if options.start_daemon:
        extra_options.append('--daemon')
    if options.stop_daemon:
        extra_options.append('--stop-daemon')
    if options.reload:
        extra_options.append('--reload')
    if options.reload_interval:
        extra_options.append('--reload-interval=%s' % options.reload_interval)
    if options.monitor_restart:
        extra_options.append('--monitor-restart')
    if options.show_status:
        extra_options.append('--status')
    if options.set_user:
        extra_options.append('--user=%s' % options.set_user)
    if options.set_group:
        extra_options.append('--group=%s' % options.set_group)
    if (options.start_daemon or
        options.stop_daemon or
        options.show_status):
        extra_options.extend(['--pid-file=%s' % options.pid_file,
                              '--log-file=%s' % options.log_file])

    paste.script.command.run(['serve', conf_file] + extra_options)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Dashboard Implementation
"""
__docformat__ = 'reStructuredText'
import zope.component
import zope.interface
from zope.viewlet import manager, viewlet
from zope.traversing.browser.absoluteurl import absoluteURL

import schooltool.person
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.dashboard import interfaces
from schooltool.securitypolicy.crowds import TeachersCrowd, StudentsCrowd


class Dashboard(manager.ViewletManagerBase):
    zope.interface.implements(interfaces.IDashboard)

    table = None

    def __getitem__(self, name):
        """See zope.interface.common.mapping.IReadMapping"""
        viewlet = super(Dashboard, self).__getitem__(name)
        if not viewlet.isAvailable():
            raise zope.component.interfaces.ComponentLookupError(
                'Viewlet `%s` is not available.' %name)

    def createTable(self):
        self.table = [[], []]
        for index in range(len(self.viewlets)):
            try:
                html = {'title':self.viewlets[index].title,
                        'content':self.viewlets[index].render()}
            except Exception, e:
                html = {'title':self.viewlets[index].title,
                        'error':"An %s error occurred: %s" %
                        (e.__class__.__name__, str(e))}
            self.table[index%2].append(html)

    def filter(self, viewlets):
        """Filter by by availability."""
        viewlets = super(Dashboard, self).filter(viewlets)
        return [(name, viewlet) for name, viewlet in viewlets
                if viewlet.isAvailable()]

    def sort(self, viewlets):
        """Sort the viewlets by weight."""
        return sorted(viewlets, key=lambda v: v[1].getWeight())

    def update(self):
        super(Dashboard, self).update()
        self.createTable()

class DashboardTabView(object):

    def dynamicVariables(self):
        return """
            var ajaxURL = '%s/dashboardContent';
            """ % (absoluteURL(ISchoolToolApplication(None), self.request))


class DashboardCategory(viewlet.ViewletBase):
    zope.interface.implements(interfaces.IDashboardCategory)

    groups = []
    weight = 100

    # See interfaces.IDashboardCategory
    title = ''

    def appURL(self):
        return absoluteURL(ISchoolToolApplication(None), self.request)

    def getWeight(self):
        """See interfaces.IDashboardCategory"""
        return int(self.weight)

    def isAvailable(self):
        return True


class PersonDashboardCategory(DashboardCategory):
    """A person-centric dashboard category.

    It is aware of person objects, and is only available for
    schooltool users.
    """

    def __init__(self, context, request, view, manager):
        super(PersonDashboardCategory, self).__init__(context, request,
                                                      view, manager)
        # Since unauthenticated users cannot be adapted to Person object,
        # we must catch the exception
        try:
            self.context = schooltool.person.interfaces.IPerson(self.request.principal)
        except TypeError:
            self.context = None

    def isAvailable(self):
        return self.context != None

class SectionsCategory(PersonDashboardCategory):
    """Dashboard area for the user's sections"""
    @property
    def isTeacher(self):
        return TeachersCrowd(self.context).contains(self.request.principal)

    @property
    def isStudent(self):
        return StudentsCrowd(self.context).contains(self.request.principal)

    def getSections(self):
        person = schooltool.person.interfaces.IPerson(self.request.principal)
        instructor_of = schooltool.course.interfaces.IInstructor(person).sections()
        for section in instructor_of:
            url = absoluteURL(section, self.request)
            url += '/gradebook'
            title = '%s %s' % (list(section.courses)[0].title, section.title)
            yield {'url': url, 'title': title}
        learner_of = schooltool.course.interfaces.ILearner(person).sections()
        for section in learner_of:
            url = absoluteURL(section, self.request)
            url += '/mygrades'
            title = '%s %s' % (list(section.courses)[0].title, section.title)
            yield {'url': url, 'title': title}



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2008 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for school years
"""
import unittest
import doctest
from datetime import date

from zope.interface.verify import verifyObject

from schooltool.term.term import Term
from schooltool.schoolyear.schoolyear import SchoolYear
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.schoolyear.interfaces import ISchoolYearContainer
from schooltool.schoolyear.testing import (setUp, tearDown,
                                           provideStubUtility,
                                           provideStubAdapter)
from schooltool.schoolyear.ftesting import schoolyear_functional_layer
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.schoolyear.schoolyear import SCHOOLYEAR_CONTAINER_KEY


def doctest_SchoolYearContainer():
    """Test for SchoolYear container

    When application starts up, the top level container should be there:

       >>> app = ISchoolToolApplication(None)
       >>> schoolyear_container = app[SCHOOLYEAR_CONTAINER_KEY]
       >>> schoolyear_container
       <schooltool.schoolyear.schoolyear.SchoolYearContainer object at ...>

       >>> verifyObject(ISchoolYearContainer, schoolyear_container)
       True

    __parent__ should be set properly:

       >>> schoolyear_container.__parent__ is app
       True

    We can access the container by adapting the application to
    ISchoolYearContainer:

       >>> ISchoolYearContainer(app) is schoolyear_container
       True

    """


def doctest_SchoolYearContainer_schoolyear_activation():
    """Test for school year activation in school year container

       >>> app = ISchoolToolApplication(None)
       >>> syc = ISchoolYearContainer(app)

    When we have no school years both active school year and next
    school year should return None:

       >>> syc.getActiveSchoolYear() is None
       True

       >>> syc.getNextSchoolYear() is None
       True

    active_id should be None as well

       >>> syc.active_id is None
       True

    but if we add a school year (it does not matter when it starts)

        >>> sy1 = SchoolYear("2005-2006", date(2005, 9, 1), date(2006, 7, 15))
        >>> syc['2005-2006'] = sy1

    it automatically becomes active

        >>> syc.getActiveSchoolYear() is sy1
        True
        >>> syc.active_id
        '2005-2006'

    though - next school year is still None

        >>> syc.getNextSchoolYear() is None
        True

    if we will add another school year, one before the active one, it
    will not get marked as the next school year

        >>> sy0 = SchoolYear("2004-2005", date(2004, 9, 1), date(2005, 7, 15))
        >>> syc['2004-2005'] = sy0

        >>> syc.getNextSchoolYear() is None
        True

    next schol year must come *after* the active one

        >>> sy2 = SchoolYear("2006-2007", date(2006, 9, 1), date(2007, 7, 15))
        >>> syc['2006-2007'] = sy2

        >>> syc.getNextSchoolYear() is sy2
        True

    we can't modify the active_id by hand

        >>> syc.active_id = '2004-2005'
        Traceback (most recent call last):
        ...
        AttributeError: can't set attribute

    we must use the activateNextSchoolYear function that automatically
    activates the next school year

        >>> syc.activateNextSchoolYear()
        >>> syc.getActiveSchoolYear() is sy2
        True
        >>> syc.getNextSchoolYear() is None
        True

    You can pass a schoolyear key to override the "default" next.

        >>> syc.activateNextSchoolYear(year_id=sy1.__name__)
        >>> syc.getActiveSchoolYear() is sy1
        True
        >>> syc.getNextSchoolYear() is sy2
        True

    """


def doctest_SchoolYear():
    """Test for SchoolYear

    School years are term like objects, that are kept in the
    schoolyear container. To construct one you must pass a title and 2
    dates to it's constructor:

        >>> sy = SchoolYear("2005-2006", date(2005, 9, 1), date(2006, 7, 15))

    Title, first and last date of the school year get set:

        >>> sy.title, sy.first, sy.last
        ('2005-2006', datetime.date(2005, 9, 1), datetime.date(2006, 7, 15))

    All school years implement ISchoolYear interface:

        >>> verifyObject(ISchoolYear, sy)
        True

    School years are kept in the schoolyear container:

        >>> syc = ISchoolYearContainer(ISchoolToolApplication(None))
        >>> syc['2005-2006'] = sy

    """


def doctest_SchoolYear_first_before_last():
    """Tests for SchoolYear

    The first day of a schoolyear must come before the last day.

        >>> sy = SchoolYear("2005-2006", date(2005, 9, 1), date(2005, 7, 15))
        Traceback (most recent call last):
        ...
        ValueError: Last date datetime.date(2005, 7, 15) less than
        first date datetime.date(2005, 9, 1)

   We cannot trick that constraint by sneakily trying to change the
   attributes after creation

        >>> sy = SchoolYear("2005-2006", date(2005, 9, 1), date(2006, 7, 15))
        >>> sy.first = date(2006, 9, 1)
        Traceback (most recent call last):
        ...
        ValueError: Last date datetime.date(2006, 7, 15) less than
        first date datetime.date(2006, 9, 1)

        >>> sy.first
        datetime.date(2005, 9, 1)

        >>> sy.last = date(2005, 7, 1)
        Traceback (most recent call last):
        ...
        ValueError: Last date datetime.date(2005, 7, 1) less than
        first date datetime.date(2005, 9, 1)

        >>> sy.last
        datetime.date(2006, 7, 15)

    One-day schoolyears are valid, albeit silly

        >>> sy = SchoolYear("2005-2006", date(2005, 9, 1), date(2005, 9, 1))

    We should still be able to set the same dates through first and last:

        >>> sy.first = date(2005, 9, 1)

        >>> sy.last = date(2005, 9, 1)

    Expanding the school year should work as well:

        >>> sy.first = date(2004, 9, 1)

        >>> sy.last = date(2006, 9, 1)

    """


def doctest_SchoolYearContainer_years_must_not_overlap():
    """Test for SchoolYearContainer and SchoolYear interaction

    School years container is for keeping school years:

        >>> syc = ISchoolYearContainer(ISchoolToolApplication(None))
        >>> sy1 = SchoolYear("2005-2006", date(2005, 9, 1), date(2006, 7, 15))
        >>> syc['2005-2006'] = sy1

        >>> sy2 = SchoolYear("2006-2007", date(2006, 9, 1), date(2007, 7, 15))
        >>> syc['2006-2007'] = sy2

    Even though you can create a schoolyear that overlaps other
    schoolyears:

        >>> sy3 = SchoolYear("2005-2007", date(2005, 9, 1), date(2007, 1, 12))

    You can't add it to the SchoolYearContainer :

        >>> syc['2005-2007'] = sy3
        Traceback (most recent call last):
        ...
        SchoolYearOverlapError: SchoolYear '2005-2007' overlaps with
          SchoolYear(s) (2005-2006, 2006-2007)

    You can't work around the restriction by changing schoolyears that
    are already in the container either:

        >>> sy1.last = date(2007, 1, 12)
        Traceback (most recent call last):
        ...
        SchoolYearOverlapError: SchoolYear '2005-2006' overlaps
        with SchoolYear(s) (2006-2007)

        >>> sy1.last
        datetime.date(2006, 7, 15)

        >>> sy2.first = date(2005, 9, 1)
        Traceback (most recent call last):
        ...
        SchoolYearOverlapError: SchoolYear '2006-2007' overlaps
        with SchoolYear(s) (2005-2006)

        >>> sy2.first
        datetime.date(2006, 9, 1)

    """


def doctest_SchoolYear_terms_must_not_overlap():
    """Test for school years with terms in them

    School years can have terms in them:

        >>> syc = ISchoolYearContainer(ISchoolToolApplication(None))
        >>> sy = SchoolYear("2005-2006", date(2005, 9, 1), date(2006, 7, 15))
        >>> syc['2005-2006'] = sy

        >>> fall = sy['fall'] = Term("Fall", date(2005, 9, 1), date(2006, 1, 12))
        >>> spring = sy['spring'] = Term("Spring", date(2006, 1, 13), date(2006, 7, 15))

    Terms can't overlap:

        >>> syc['2005-2006']['fall_2'] = Term("Fall 2", date(2005, 9, 1), date(2006, 1, 12))
        Traceback (most recent call last):
        ...
        TermOverlapError: Date range you have selected overlaps with term(s) (Fall)

    You can't work around the restriction by changing terms
    directly either:

        >>> fall.last = date(2006, 1, 17)
        Traceback (most recent call last):
        ...
        TermOverlapError: Date range you have selected overlaps with term(s) (Spring)

        >>> fall.last
        datetime.date(2006, 1, 12)

        >>> spring.first = date(2005, 10, 15)
        Traceback (most recent call last):
        ...
        TermOverlapError: Date range you have selected overlaps with term(s) (Fall)

        >>> spring.first
        datetime.date(2006, 1, 13)

    """


def doctest_SchoolYear_terms_must_be_fully_contained():
    """Test for Term and SchoolYear interaction

    School years can have terms in them:

        >>> syc = ISchoolYearContainer(ISchoolToolApplication(None))
        >>> sy = SchoolYear("2005-2006", date(2005, 9, 1), date(2006, 7, 15))
        >>> syc['2005-2006'] = sy

        >>> sy['fall'] = Term("Fall", date(2005, 9, 1), date(2006, 1, 12))
        >>> sy['spring'] = Term("Spring", date(2006, 1, 13), date(2006, 7, 15))

    But term dates are limited to the ones that are in the range of
    the school year:

        >>> sy['fall'] = Term("Fall", date(2006, 8, 31), date(2007, 1, 12))
        Traceback (most recent call last):
        ...
        ValueError: Term can't end after the school year ends!

        >>> sy['fall'] = Term("Fall", date(2006, 9, 1), date(2007, 7, 16))
        Traceback (most recent call last):
        ...
        ValueError: Term can't end after the school year ends!

    You can't work around these limmitations by modifying the term
    after it was added:

        >>> sy['fall'].first = date(2005, 8, 1)
        Traceback (most recent call last):
        ...
        TermDateNotInSchoolYear: ...

    Also you should be unable to shrink the school year without
    changing boundaries of the terms first:

        >>> sy.first = date(2005, 9, 2)
        Traceback (most recent call last):
        ...
        TermOverflowError: Date range you are trying to set is too small to
        contain following term(s) (Fall)

        >>> sy.last = date(2006, 6, 15)
        Traceback (most recent call last):
        ...
        TermOverflowError: Date range you are trying to set is too small to
        contain following term(s) (Spring)

    """


def test_suite():
    optionflags = doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS
    suite = doctest.DocTestSuite(optionflags=optionflags,
                                 extraglobs={'provideAdapter': provideStubAdapter,
                                             'provideUtility': provideStubUtility},
                                 setUp=setUp, tearDown=tearDown)
    suite.layer = schoolyear_functional_layer
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2007 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Resource Booking caledar and events
"""
from zope.interface import implements
from zope.component import queryAdapter
from zope.location.location import Location, locate
from zope.publisher.interfaces import NotFound
from zope.session.interfaces import ISession

from schooltool.calendar.simple import ImmutableCalendar
from schooltool.resource.interfaces import IBookingCalendar
from schooltool.calendar.simple import SimpleCalendarEvent
from schooltool.resource.interfaces import IBookingCalendarEvent
from schooltool.resource.interfaces import IBookingTimetableEvent
from schooltool.app.interfaces import ISchoolToolCalendar
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.person.interfaces import IPerson
from schooltool.timetable.interfaces import ITimetableContainer
from schooltool.timetable.calendar import ImmutableScheduleCalendar
from schooltool.traverser.traverser import TraverserPlugin


class BookingCalendarEvent(SimpleCalendarEvent):
    implements(IBookingCalendarEvent)


class BookingTimetableEvent(SimpleCalendarEvent):
    implements(IBookingTimetableEvent)


def createBookingCalendar(calendar, calendars, event_factory=BookingCalendarEvent):
    events = []
    for event in calendar:
        resources = []
        for resource_calendar in calendars:
            evts = list(resource_calendar.expand(event.dtstart,
                                                 event.dtstart + event.duration))
            if not list(evts):
                resources.append(resource_calendar.__parent__)

        if resources:
            event = event_factory(event.dtstart, event.duration,
                                  event.title,
                                  description=event.description,
                                  unique_id=event.unique_id)
            event.resources = resources
            event.__parent__ = None
            events.append(event)

    return ImmutableCalendar(events)


def getSelectedResourceCalendars(request):
    rc = ISchoolToolApplication(None)['resources']
    session = ISession(request)['schooltool.resource']
    resource_calendars = [ISchoolToolCalendar(rc[resource_id])
                          for resource_id in session['bookingSelection']]
    return resource_calendars


class ResourceBookingCalendar(ImmutableCalendar, Location):
    implements(IBookingCalendar)

    def __init__(self, context):
        self.context = context
        self.__parent__ = self.context
        self.__name__ = 'booking'
        self.title = "Booking Calendar"

    def expand(self, start, end):
        app = ISchoolToolApplication(None)

        school_timetables = ITimetableContainer(app, None)

        if school_timetables is None or school_timetables.default is None:
            return []

        calendar = ImmutableScheduleCalendar(school_timetables.default)
        events = []
        events.extend(calendar.expand(start, end))

        timetable_calendar = ImmutableCalendar(events)

        resource_calendars = getSelectedResourceCalendars(self.request)
        booking_calendar = createBookingCalendar(timetable_calendar,
                                                 resource_calendars,
                                                 event_factory=BookingTimetableEvent)
        return booking_calendar.expand(start, end)

    def __iter__(self):
        return iter([])


class ResourceBookingTraverserPlugin(TraverserPlugin):
    """Traverse to an adapter by name."""

    def traverse(self, name):
        bookingCalendar = queryAdapter(self.context, IBookingCalendar, name='')
        if bookingCalendar is None:
            raise NotFound(self.context, name, self.request)
        bookingCalendar.request = self.request
        return bookingCalendar


class ResourceBookingCalendarProvider(object):

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def getCalendars(self):
        """Get a list of calendars to display.

        Yields tuples (calendar, color1, color2).
        """
        person_calendar = ISchoolToolCalendar(IPerson(self.request.principal))
        resource_calendars = getSelectedResourceCalendars(self.request)
        booking_calendar = createBookingCalendar(person_calendar,
                                                 resource_calendars)
        yield (booking_calendar, '#9db8d2', '#7590ae')
        yield (self.context, '#bfdaf4', '#99b2ce')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool Application
"""

def registerTestSetup():
    from zope.interface import classImplements
    from schooltool.testing import registry

    def haveCalendar():
        from schooltool.app.app import SchoolToolApplication
        from schooltool.app.interfaces import IHaveCalendar
        if not IHaveCalendar.implementedBy(SchoolToolApplication):
            classImplements(SchoolToolApplication, IHaveCalendar)
    registry.register('CalendarComponents', haveCalendar)

    def haveTimetables():
        from schooltool.timetable.interfaces import IHaveTimetables
        from schooltool.timetable.interfaces import IHaveSchedule

        from schooltool.schoolyear.schoolyear import SchoolYear
        if not IHaveTimetables.implementedBy(SchoolYear):
            classImplements(SchoolYear, IHaveTimetables)

        from schooltool.course.section import Section
        if not IHaveTimetables.implementedBy(Section):
            classImplements(Section, IHaveSchedule)
    registry.register('TimetablesComponents', haveTimetables)

registerTestSetup()
del registerTestSetup



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Recurrence rules.
"""

import datetime
import calendar

from zope.interface import implements

from schooltool.calendar.icalendar import ical_weekdays
from schooltool.calendar.icalendar import ical_date, ical_datetime
from schooltool.calendar.interfaces import Unchanged
from schooltool.calendar.interfaces import \
    IDailyRecurrenceRule, IWeeklyRecurrenceRule, IMonthlyRecurrenceRule, \
    IYearlyRecurrenceRule, IRecurrenceRule


class RecurrenceRule(object):
    """An abstract base class that implements common behaviour."""

    implements(IRecurrenceRule)

    interval = property(lambda self: self._interval)
    count = property(lambda self: self._count)
    until = property(lambda self: self._until)
    exceptions = property(lambda self: self._exceptions)

    # A string that represents the recurrence frequency in iCalendar.
    # Must be overridden by subclasses.
    ical_freq = None

    def __init__(self, interval=1, count=None, until=None, exceptions=()):
        self._interval = interval
        self._count = count
        if until and isinstance(until, datetime.datetime):
            self._until = until.date()
        else:
            self._until = until
        self._exceptions = tuple(exceptions) # XXX I'd prefer a set.
        self._validate()

    def _validate(self):
        if self.count is not None and self.until is not None:
            raise ValueError("count and until cannot be both set (%s, %s)"
                             % (self.count, self.until))
        if not self.interval >= 1:
            raise ValueError("interval must be a positive integer (got %r)"
                             % (self.interval, ))
        for ex in self.exceptions:
            if not isinstance(ex, datetime.date):
                raise ValueError("Exceptions must be a sequence of"
                                 " datetime.dates (got %r in exceptions)"
                                 % (ex, ))

    def replace(self, interval=Unchanged, count=Unchanged, until=Unchanged,
                exceptions=Unchanged):
        if interval is Unchanged:
            interval = self.interval
        if count is Unchanged:
            count = self.count
        if until is Unchanged:
            until = self.until
        if exceptions is Unchanged:
            exceptions = self.exceptions
        return self.__class__(interval, count, until, exceptions)

    def __repr__(self):
        return '%s(%r, %r, %r, %r)' % (self.__class__.__name__, self.interval,
                                       self.count, self.until,
                                       self.exceptions)

    def _tupleForComparison(self):
        return (self.__class__.__name__, self.interval, self.count,
                self.until, tuple(self.exceptions))

    def __eq__(self, other):
        """See if self == other."""
        if isinstance(other, RecurrenceRule):
            return self._tupleForComparison() == other._tupleForComparison()
        else:
            return False

    def __ne__(self, other):
        """See if self != other."""
        return not self == other

    def __hash__(self):
        """Return the hash value of this recurrence rule.

        It is guaranteed that if recurrence rules compare equal, hash will
        return the same value.
        """
        return hash(self._tupleForComparison())

    def apply(self, event, startdate=None, enddate=None):
        """Generator that generates dates of recurrences"""
        if enddate:
            assert isinstance(enddate, datetime.date), "enddate must be a date"
            assert not isinstance(enddate, datetime.datetime), \
                    "enddate must be a date, not a datetime"
        if startdate is None:
            startdate = event.dtstart.date()

        count, cur = self._scroll(event, startdate)

        while True:
            if ((enddate and cur > enddate) or
                (self.count is not None and count >= self.count) or
                (self.until and cur > self.until)):
                break
            if cur not in self.exceptions and cur >= startdate:
                yield cur
            count += 1
            cur = self._nextRecurrence(cur)

    def _nextRecurrence(self, date):
        """Add the basic step of recurrence to the date."""
        return date + self.interval * date.resolution

    def iCalRepresentation(self, dtstart):
        """See IRecurrenceRule"""
        assert self.ical_freq, 'RecurrenceRule.ical_freq must be overridden'

        if self.count:
            args = 'COUNT=%d;' % self.count
        elif self.until:
            args = 'UNTIL=%s;' % ical_datetime(self.until)
        else:
            args = ''
        extra_args = self._iCalArgs(dtstart)
        if extra_args is not None:
            args += extra_args + ';'

        result = ['RRULE:FREQ=%s;%sINTERVAL=%d'
                  % (self.ical_freq, args, self.interval)]

        if self.exceptions:
            # Exceptions should include the exact time portion as well
            # (this was implemented in revision 1860), however,
            # Mozilla Calendar refuses to work with such exceptions.
            dates = ','.join([ical_date(d) for d in self.exceptions])
            result.append('EXDATE;VALUE=DATE:' + dates)
        return result

    def _iCalArgs(self, dtstart):
        """Return extra iCal arguments as a string.

        Should be overridden by child classes that have specific arguments.
        The returned string must not include the semicolon separator.
        If None is returned, no arguments are inserted.
        """
        pass


class DailyRecurrenceRule(RecurrenceRule):
    """Daily recurrence rule.

    Immutable hashable object.
    """
    implements(IDailyRecurrenceRule)

    ical_freq = 'DAILY'

    def _scroll(self, event, startdate):
        """Given a startdate, finds a nearest event recurrence and its nr"""
        if startdate >= event.dtstart.date():
            days = (startdate - event.dtstart.date()).days
        else:
            days = 0
        count = days / self.interval
        cur = event.dtstart.date() + datetime.timedelta(count * self.interval)
        return count, cur



class YearlyRecurrenceRule(RecurrenceRule):
    """Yearly recurrence rule.

    Immutable hashable object.
    """
    implements(IYearlyRecurrenceRule)

    ical_freq = 'YEARLY'

    def _nextRecurrence(self, date):
        """Adds the basic step of recurrence to the date"""
        delta = self.interval
        while True:
            try:
                nextyear = date.year + delta
                return date.replace(year=nextyear)
            except ValueError:
                # Jump over illegal leap days
                delta += self.interval

    def _scroll(self, event, startdate):
        """Given a startdate, finds a nearest event recurrence and its nr"""
        if startdate >= event.dtstart.date():
            years = startdate.year - event.dtstart.date().year
        else:
            years = 0
        count = years / self.interval
        while True:
            try:
                newyear = event.dtstart.date().year + count * self.interval
                cur = event.dtstart.date().replace(year=newyear)
                return count, cur
            except ValueError:
                count -= self.interval

    def _iCalArgs(self, dtstart):
        """Return iCalendar parameters specific to monthly reccurence."""
        # KOrganizer wants explicit BYMONTH and BYMONTHDAY arguments.
        # Maybe it is a good idea to add them for the sake of explicitness.


class WeeklyRecurrenceRule(RecurrenceRule):
    """Weekly recurrence rule."""

    implements(IWeeklyRecurrenceRule)

    weekdays = property(lambda self: self._weekdays)

    ical_freq = 'WEEKLY'

    def __init__(self, interval=1, count=None, until=None, exceptions=(),
                 weekdays=()):
        self._interval = interval
        self._count = count
        if until and isinstance(until, datetime.datetime):
            self._until = until.date()
        else:
            self._until = until
        self._exceptions = tuple(exceptions)
        self._weekdays = tuple(weekdays)
        self._validate()

    def __repr__(self):
        return '%s(%r, %r, %r, %r, %r)' % (
            self.__class__.__name__, self.interval,
            self.count, self.until, self.exceptions, self.weekdays)

    def _validate(self):
        RecurrenceRule._validate(self)
        for dow in self.weekdays:
            if not isinstance(dow, int) or not 0 <= dow <= 6:
                raise ValueError("Day of week must be an integer 0..6 (got %r)"
                                 % (dow, ))

    def replace(self, interval=Unchanged, count=Unchanged, until=Unchanged,
                exceptions=Unchanged, weekdays=Unchanged, monthly=Unchanged):
        if interval is Unchanged:
            interval = self.interval
        if count is Unchanged:
            count = self.count
        if until is Unchanged:
            until = self.until
        if exceptions is Unchanged:
            exceptions = self.exceptions
        if weekdays is Unchanged:
            weekdays = self.weekdays
        return self.__class__(interval, count, until, exceptions, weekdays)

    def _tupleForComparison(self):
        return (self.__class__.__name__, self.interval, self.count,
                self.until, self.exceptions, self.weekdays)

    def _scroll(self, event, startdate):
        """Fast-forward to recurrences near startdate"""
        start = event.dtstart.date()
        if startdate >= start:
            weeks = weekspan(start, startdate)
        else:
            weeks = 0
        perweek = len(self.weekdays) + 1 # nr of recurrences per week
        count = weeks / self.interval * perweek
        if count > perweek:
            count -= perweek # to get entries with weekday < that of start
        offset = datetime.timedelta(7) * count / perweek * self.interval
        return count, start + offset

    def apply(self, event, startdate=None, enddate=None):
        """Generate dates of recurrences."""
        if enddate:
            assert isinstance(enddate, datetime.date), "enddate must be a date"
            assert not isinstance(enddate, datetime.datetime), \
                    "enddate must be a date, not a datetime"
        start = event.dtstart.date()
        if startdate is None:
            startdate = start
        count, cur = self._scroll(event, startdate)
        weekdays = set(self.weekdays)
        weekdays.add(event.dtstart.weekday())
        while True:
            if ((enddate and cur > enddate) or
                (self.count is not None and count >= self.count) or
                (self.until and cur > self.until)):
                break
            # Check that this is the correct week and
            # the desired weekday
            if (weekspan(start, cur) % self.interval == 0 and
                cur.weekday() in weekdays):
                if cur not in self.exceptions and cur >= startdate:
                    yield cur
                count += 1
            cur = self._nextRecurrence(cur)

    def _nextRecurrence(self, date):
        """Add the basic step of recurrence to the date."""
        return date + date.resolution

    def _iCalArgs(self, dtstart):
        """Return iCalendar parameters specific to weekly reccurence."""
        if dtstart is not None:
            weekdays = set(self.weekdays)
            weekdays.add(dtstart.weekday())
            return 'BYDAY=' + ','.join([ical_weekdays[weekday]
                                        for weekday in weekdays])


class MonthlyRecurrenceRule(RecurrenceRule):
    """Monthly recurrence rule.

    Immutable hashable object.
    """
    implements(IMonthlyRecurrenceRule)

    monthly = property(lambda self: self._monthly)

    ical_freq = 'MONTHLY'

    def __init__(self, interval=1, count=None, until=None, exceptions=(),
                 monthly="monthday"):
        self._interval = interval
        self._count = count
        if until and isinstance(until, datetime.datetime):
            self._until = until.date()
        else:
            self._until = until
        self._exceptions = tuple(exceptions)
        self._monthly = monthly
        self._validate()

    def __repr__(self):
        return '%s(%r, %r, %r, %r, %r)' % (
            self.__class__.__name__, self.interval,
            self.count, self.until, self.exceptions, self.monthly)

    def _validate(self):
        RecurrenceRule._validate(self)
        if self.monthly not in ("monthday", "weekday", "lastweekday"):
            raise ValueError("monthly must be one of 'monthday', 'weekday',"
                             " 'lastweekday'. Got %r" % (self.monthly, ))

    def replace(self, interval=Unchanged, count=Unchanged, until=Unchanged,
                exceptions=Unchanged, weekdays=Unchanged, monthly=Unchanged):
        if interval is Unchanged:
            interval = self.interval
        if count is Unchanged:
            count = self.count
        if until is Unchanged:
            until = self.until
        if exceptions is Unchanged:
            exceptions = tuple(self.exceptions)
        if monthly is Unchanged:
            monthly = self.monthly
        return self.__class__(interval, count, until, exceptions, monthly)

    def _tupleForComparison(self):
        return (self.__class__.__name__, self.interval, self.count,
                self.until, self.exceptions, self.monthly)

    def _scroll(self, event, startdate):
        """Given a startdate, finds a nearest event recurrence and its nr"""
        dstart = event.dtstart.date()
        if startdate > event.dtstart.date():
            months = (12 * (startdate.year - dstart.year) +
                      startdate.month - dstart.month)
        else:
            months = 0

        count = months / self.interval
        while True:
            try:
                year, month = divmod(dstart.year * 12 + dstart.month +
                                     count * self.interval - 1, 12)
                month += 1 # 1..12
                cur = event.dtstart.date().replace(year=year, month=month)
                return count, cur
            except ValueError:
                count -= 1

    def _nextRecurrence(self, date):
        """Add basic step of recurrence to the date."""
        year = date.year
        month = date.month
        while True:
            year, month = divmod(year * 12 + month - 1 + self.interval, 12)
            month += 1 # convert 0..11 to 1..12
            try:
                return date.replace(year=year, month=month)
            except ValueError:
                continue

    def apply(self, event, startdate=None, enddate=None):
        if enddate:
            assert isinstance(enddate, datetime.date), "enddate must be a date"
            assert not isinstance(enddate, datetime.datetime), \
                    "enddate must be a date, not a datetime"
        if self.monthly == 'monthday':
            for date in  RecurrenceRule.apply(self, event, startdate, enddate):
                yield date
        elif self.monthly == 'weekday':
            for date in self._applyWeekday(event, startdate, enddate):
                yield date
        elif self.monthly == 'lastweekday':
            for date in self._applyLastWeekday(event, startdate, enddate):
                yield date

    def _applyWeekday(self, event, startdate=None, enddate=None):
        """Generator that generates dates of recurrences."""
        start = event.dtstart.date()
        weekday = start.weekday()
        # Which week of the month is it
        index = (start.day - 1) / 7 + 1

        if startdate is None:
            startdate = start

        count, cur = self._scroll(event, startdate)
        year = cur.year
        month = cur.month

        while True:
            cur = monthindex(year, month, index, weekday)
            if ((enddate and cur > enddate) or
                (self.count is not None and count >= self.count) or
                (self.until and cur > self.until)):
                break
            if cur not in self.exceptions and cur >= startdate:
                yield cur
            count += 1
            # Next month, please.
            year, month = divmod(year * 12 + month + self.interval - 1, 12)
            month += 1

    def _applyLastWeekday(self, event, startdate=None, enddate=None):
        """Generator that generates dates of recurrences."""
        start = event.dtstart.date()
        weekday = start.weekday()
        daysinmonth = calendar.monthrange(start.year, start.month)[1]
        index = (start.day - daysinmonth - 1) / 7

        if startdate is None:
            startdate = start

        count, cur = self._scroll(event, startdate)
        year = cur.year
        month = cur.month

        while True:
            cur = monthindex(year, month, index, weekday)
            if ((enddate and cur > enddate) or
                (self.count is not None and count >= self.count) or
                (self.until and cur > self.until)):
                break
            if cur not in self.exceptions and cur >= startdate:
                yield cur
            count += 1
            # Next month, please.
            year, month = divmod(year * 12 + month + self.interval - 1, 12)
            month += 1

    def _iCalArgs(self, dtstart):
        """Return iCalendar parameters specific to monthly reccurence."""
        if self.monthly == 'monthday':
            return 'BYMONTHDAY=%d' % dtstart.day
        elif self.monthly == 'weekday':
            week = dtstart.day / 7 + 1
            return 'BYDAY=%d%s' % (week, ical_weekdays[dtstart.weekday()])
        elif self.monthly == 'lastweekday':
            return 'BYDAY=-1%s' % ical_weekdays[dtstart.weekday()]
        else:
            raise NotImplementedError(self.monthly)


#
# Calendaring functions
#


def weekspan(first, second):
    """Return the distance in weeks between dates.

    For days in the same ISO week, the result is 0.
    For days in adjacent weeks, it is 1, etc.
    """
    firstmonday = first - datetime.timedelta(first.weekday())
    secondmonday = second - datetime.timedelta(second.weekday())
    return (secondmonday - firstmonday).days / 7


def monthindex(year, month, index, weekday):
    """Return the (index)th weekday of the month in a year.

    May return a date beyond month if index is too big.
    """
    # make corrections for the negative index
    # if index is negative, we're really interested in the next month's
    # first weekday, minus n weeks
    if index < 0:
        yeardelta, month = divmod(month, 12)
        year += yeardelta
        month += 1
        index += 1

    # find first weekday
    for day in range(1, 8):
        if datetime.date(year, month, day).weekday() == weekday:
            break

    # calculate the timedelta to the index-th
    shift = (index - 1) * datetime.timedelta(7)

    # return the result
    return datetime.date(year, month, day) + shift



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.resource.sampledata
"""

import unittest
import doctest

from zope.interface.verify import verifyObject
from zope.app.testing import setup

from schooltool.testing import setup as stsetup


def setUp(test):
    setup.placefulSetUp()


def tearDown(test):
    setup.placefulTearDown()


def doctest_SampleResources():
    """A sample data plugin that generates resources

        >>> from schooltool.resource.interfaces import ILocation
        >>> from schooltool.resource.sampledata import SampleResources
        >>> from schooltool.sampledata.interfaces import ISampleDataPlugin
        >>> plugin = SampleResources()
        >>> verifyObject(ISampleDataPlugin, plugin)
        True

        >>> app = stsetup.setUpSchoolToolSite()
        >>> len(app['resources'])
        0

        >>> plugin.generate(app, 42)

        >>> len(app['resources'])
        88

        >>> for i in range(64):
        ...     location = app['resources']['room%02d' % i]
        ...     assert ILocation.providedBy(location)
        ...     assert location.type == 'Classroom'

        >>> for i in range(24):
        ...     projector = app['resources']['projector%02d' % i]
        ...     assert not ILocation.providedBy(projector)
        ...     assert projector.type == "Projector"


    """


def test_suite():
    return unittest.TestSuite([
        doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                             optionflags=doctest.ELLIPSIS),
        ])


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for schooltool views.
"""
import unittest
import doctest

from zope.interface import implements
from zope.component import provideAdapter
from zope.interface.verify import verifyObject
from zope.location.interfaces import ILocation
from zope.app.testing import setup
from zope.annotation.interfaces import IAttributeAnnotatable
from zope.app.dependable.interfaces import IDependable
from zope.i18n import translate
from zope.publisher.browser import TestRequest
from zope.traversing.interfaces import ITraversable, IPathAdapter

from schooltool.app.browser.testing import setUp, tearDown
from schooltool.testing import setup as sbsetup


class SomeAnnotatable(object):
    implements(IAttributeAnnotatable)


def doctest_SchoolToolAPI():
    r"""Tests for SchoolToolAPI.

        >>> from zope.tales.interfaces import ITALESFunctionNamespace
        >>> from schooltool.app.browser import SchoolToolAPI
        >>> context = object()
        >>> ns = SchoolToolAPI(context)
        >>> verifyObject(ITALESFunctionNamespace, ns)
        True

    """


def doctest_SchoolToolAPI_app():
    r"""Tests for SchoolToolAPI.app.

    'context/schooltool:app' returns the nearest ISchoolToolApplication

        >>> app = sbsetup.setUpSchoolToolSite()

        >>> from schooltool.app.browser import SchoolToolAPI
        >>> SchoolToolAPI(app['persons']).app is app
        True

    It does so even for objects that do not adapt to ISchoolToolApplication,
    but are sublocations of SchoolToolApplication:

        >>> class Adding:
        ...     implements(ILocation)
        ...     __parent__ = app['persons']
        ...
        >>> adding = Adding()
        >>> SchoolToolAPI(adding).app is app
        True

    """


def doctest_SchoolToolAPI_content():
    r"""Tests for SchoolToolAPI.content.

        >>> from zope.traversing.interfaces import ITraversable
        >>> from zope.tales.interfaces import ITALESFunctionNamespace

        >>> from schooltool.app.browser import SchoolToolAPI
        >>> from schooltool.skin.flourish.interfaces import IContentProviders

    'context/schooltool:content' returns a traversable to content providers
    for given context and request, view defined in tal.

        >>> class TALContext(object):
        ...     def __init__(self, vars=()):
        ...         self.vars = dict(vars)

        >>> application = sbsetup.setUpSchoolToolSite()

    We need the api.

        >>> api = SchoolToolAPI(application['persons'])

        >>> print api.content
        None

    Tal engine must be set.  It will be set externally, because:

        >>> verifyObject(ITALESFunctionNamespace, api)
        True

    But in this test we'll do that manually.

        >>> api.setEngine(TALContext())

        >>> print api.content
        None

    Content providers are looked up via adapter.

        >>> class ContentProvidersStub(object):
        ...     implements(IContentProviders)
        ...     def __init__(self, context, request, view):
        ...         self.context = context
        ...         self.request = request
        ...         self.view = view

        >>> issubclass(IContentProviders, ITraversable)
        True

        >>> class ViewStub(object):
        ...     def __init__(self, context, request):
        ...         self.context = context
        ...         self.request = request

        >>> view = ViewStub(application['schooltool.schoolyear'], TestRequest())

        >>> provideAdapter(ContentProvidersStub,
        ...                (None, None, None),
        ...                IContentProviders)

        >>> providers = api.content
        >>> providers.context, providers.request, providers.view
        (<...PersonContainer ...>, None, None)

    Parameters for the adapter are looked up from tal variables.

        >>> api.setEngine(TALContext({
        ...     'context': view.context,
        ...     'request': view.request,
        ...     'view': view,
        ...     }))

        >>> providers = api.content
        >>> print providers
        <...ContentProvidersStub object at ...>

        >>> providers.context
        <...PersonContainer object at ...>

        >>> providers.request
        <...TestRequest ...>

        >>> providers.view
        <...ViewStub object at ...>

    Note that:

        >>> providers.context is api.context
        True

        >>> providers.context is not api.engine.vars['context']
        True

    If content providers implement ITALESFunctionNamespace, api.contacts
    also set the engine context.

        >>> class TALAwareContentProviders(ContentProvidersStub):
        ...     implements(ITALESFunctionNamespace)
        ...     engine = None
        ...     def setEngine(self, engine):
        ...         self.engine = engine

        >>> provideAdapter(TALAwareContentProviders,
        ...                (None, None, None),
        ...                IContentProviders)

        >>> providers = api.content

        >>> providers
        <...TALAwareContentProviders object at ...>

        >>> providers.engine
        <...TALContext object at ...>

    """


def doctest_SchoolToolAPI_preferences():
    r"""Tests for SchoolToolAPI.preferences.

        >>> setup.setUpAnnotations()
        >>> from schooltool.app.interfaces import ISchoolToolApplication
        >>> from schooltool.app.interfaces import IApplicationPreferences
        >>> from schooltool.app.app import getApplicationPreferences
        >>> provideAdapter(getApplicationPreferences,
        ...                (ISchoolToolApplication,), IApplicationPreferences)

    'context/schooltool:preferences' returns an ApplicationPreferences object
    for the nearest ISchoolToolApplication

        >>> from schooltool.app.browser import SchoolToolAPI
        >>> app = sbsetup.setUpSchoolToolSite()
        >>> preferences = SchoolToolAPI(app).preferences
        >>> preferences.title
        u'Your School'

        >>> preferences is SchoolToolAPI(app['persons']).preferences
        True

    """


def doctest_SchoolToolAPI_person():
    r"""Tests for SchoolToolAPI.person.

    'context/schooltool:person' adapts the context to IPerson:

        >>> from schooltool.app.browser import SchoolToolAPI
        >>> from schooltool.person.person import Person
        >>> p = Person()

        >>> SchoolToolAPI(p).person is p
        True
        >>> SchoolToolAPI(None).person is None
        True

    """


def doctest_SchoolToolAPI_authenticated():
    r"""Tests for SchoolToolAPI.authenticated.

    'context/schooltool:authenticated' checks whether context is an
    authenticated principal

        >>> from zope.principalregistry.principalregistry \
        ...     import Principal, UnauthenticatedPrincipal
        >>> root = Principal('root', 'Admin', 'Supreme user', 'root', 'secret')
        >>> anonymous = UnauthenticatedPrincipal('anonymous', 'Anonymous',
        ...                             "Anyone who did not bother to log in")

        >>> from schooltool.app.browser import SchoolToolAPI
        >>> SchoolToolAPI(root).authenticated
        True
        >>> SchoolToolAPI(anonymous).authenticated
        False

        >>> from schooltool.person.person import Person
        >>> SchoolToolAPI(Person()).authenticated
        ... # doctest: +NORMALIZE_WHITESPACE, +ELLIPSIS
        Traceback (most recent call last):
          ...
        TypeError: schooltool:authenticated can only be applied to a principal
        but was applied on <schooltool.person.person.Person object at ...>

    """


def doctest_SchoolToolAPI_has_dependents():
    r"""Tests for SchoolToolAPI.has_dependents.

        >>> from schooltool.app.browser import SchoolToolAPI

        >>> obj = object()
        >>> SchoolToolAPI(obj).has_dependents
        False

        >>> obj = SomeAnnotatable()
        >>> SchoolToolAPI(obj).has_dependents
        False

        >>> IDependable(obj).addDependent('foo')
        >>> SchoolToolAPI(obj).has_dependents
        True

    """


def doctest_SortBy():
    """Tests for SortBy adapter.

        >>> from schooltool.app.browser import SortBy
        >>> adapter = SortBy([])

        >>> verifyObject(IPathAdapter, adapter)
        True
        >>> verifyObject(ITraversable, adapter)
        True

    You can sort empty lists

        >>> adapter.traverse('name')
        []

    You can sort lists of dicts

        >>> a_list = [{'id': 42, 'title': 'How to get ahead in navigation'},
        ...           {'id': 11, 'title': 'The ultimate answer: 6 * 9'},
        ...           {'id': 33, 'title': 'Alphabet for beginners'}]
        >>> for item in SortBy(a_list).traverse('title'):
        ...     print item['title']
        Alphabet for beginners
        How to get ahead in navigation
        The ultimate answer: 6 * 9

    You can sort lists of objects by attribute

        >>> class SomeObject:
        ...     def __init__(self, name):
        ...         self.name = name
        >>> another_list = map(SomeObject, ['orange', 'apple', 'pear'])
        >>> for item in SortBy(another_list).traverse('name'):
        ...     print item.name
        apple
        orange
        pear

    You cannot mix and match objects and dicts in the same list, though.
    Also, if you specify the name of a method, SortBy will not call that
    method to get sort keys.

    You can sort arbitrary iterables, in fact:

        >>> import itertools
        >>> iterable = itertools.chain(another_list, another_list)
        >>> for item in SortBy(iterable).traverse('name'):
        ...     print item.name
        apple
        apple
        orange
        orange
        pear
        pear

    """


def doctest_CanAccess():
    """Tests for CanAccess adapter.

        >>> from schooltool.app.browser import CanAccess
        >>> context = object()

    Let's create the adapter:

        >>> adapter = CanAccess(context)
        >>> adapter.context is context
        True

    The adapter's traverse() method simply calls canAccess on the context.
    Its functionality is tested in functional tests already and is too much
    trouble to unit test, so we just check the interface:

        >>> verifyObject(ITraversable, adapter)
        True
        >>> verifyObject(IPathAdapter, adapter)
        True

    """


def doctest_SortBy_security():
    """Regression test for http://issues.schooltool.org/issue174

        >>> from zope.security.management import newInteraction
        >>> from zope.security.management import endInteraction
        >>> endInteraction()
        >>> newInteraction(TestRequest())

    Fairly standard condition: a security wrapped object.

        >>> class SacredObj:
        ...     title = 'Ribbit!'
        >>> obj = SacredObj()

        >>> from zope.security.checker import NamesChecker
        >>> from zope.security.checker import ProxyFactory
        >>> checker = NamesChecker(['title'], 'schooltool.View')
        >>> protected_obj = ProxyFactory(obj, checker)
        >>> a_list = [protected_obj]

    When we sort it, we should get Unauthorized for the 'title' attribute (as
    opposed to ForbiddenAttribute for '__getitem__', which will happen if
    hasattr hides the first error).

        >>> from schooltool.app.browser import SortBy
        >>> list(SortBy(a_list).traverse('title'))  # doctest: +ELLIPSIS
        Traceback (most recent call last):
          ...
        Unauthorized: (<...SacredObj instance...>, 'title', 'schooltool.View')

    Tear down:

        >>> endInteraction()

    """

def doctest_SchoolToolSized():
    """Unit tests for SchoolToolSized.

      >>> from schooltool.app.browser import SchoolToolSized
      >>> from schooltool.person.person import Person

      >>> app = sbsetup.setUpSchoolToolSite()
      >>> sized = SchoolToolSized(app)

      >>> sized.sizeForSorting(), translate(sized.sizeForDisplay())
      ((u'Persons', 0), u'0 persons')

      >>> persons = app['persons']
      >>> persons['gintas'] = Person(u'gintas')

      >>> sized.sizeForSorting(), translate(sized.sizeForDisplay())
      ((u'Persons', 1), u'1 person')

      >>> persons['ignas'] = Person(u'ignas')
      >>> persons['marius'] = Person(u'marius')

      >>> sized.sizeForSorting(), translate(sized.sizeForDisplay())
      ((u'Persons', 3), u'3 persons')

    """


def doctest_ViewPrefences():
    r"""Unit tests for ViewPreferences.

        >>> from schooltool.app.browser import ViewPreferences
        >>> from schooltool.person.interfaces import IPerson
        >>> from schooltool.person.interfaces import IPersonPreferences
        >>> import calendar
        >>> class PreferenceStub:
        ...     weekstart = calendar.MONDAY
        ...     timeformat = "%H:%M"
        ...     dateformat = "%Y-%m-%d"
        ...     timezone = 'UTC'
        >>> class PersonStub:
        ...     def __conform__(self, interface):
        ...         if interface is IPersonPreferences:
        ...             return PreferenceStub()
        >>> class PrincipalStub:
        ...     def __conform__(self, interface):
        ...         if interface is IPerson:
        ...             return PersonStub()
        >>> request = TestRequest()
        >>> request.setPrincipal(PrincipalStub())
        >>> prefs = ViewPreferences(request)
        >>> from datetime import datetime
        >>> prefs.timezone.tzname(datetime.utcnow())
        'UTC'
        >>> prefs.timeformat
        '%H:%M'
        >>> prefs.dateformat
        '%Y-%m-%d'
        >>> prefs.first_day_of_week
        0
        >>> from pytz import utc
        >>> prefs.renderDatetime(datetime(2005, 1, 7, 14, 15, tzinfo=utc))
        '2005-01-07 14:15'

    We have no principal (anonymous user) and no SchoolTool site (test
    environment):

        >>> request = TestRequest()
        >>> prefs = ViewPreferences(request)
        >>> from datetime import datetime
        >>> prefs.timezone.tzname(datetime.utcnow())
        'UTC'
        >>> prefs.timeformat
        '%H:%M'
        >>> prefs.dateformat
        '%Y-%m-%d'
        >>> prefs.first_day_of_week
        0

    Let's set up a SchoolTool site:

        >>> setup.setUpAnnotations()
        >>> from schooltool.app.app import getApplicationPreferences
        >>> app = sbsetup.setUpSchoolToolSite()
        >>> from schooltool.app.interfaces import IApplicationPreferences
        >>> from schooltool.app.interfaces import ISchoolToolApplication
        >>> provideAdapter(getApplicationPreferences,
        ...                (ISchoolToolApplication,), IApplicationPreferences)
        >>> aprefs = IApplicationPreferences(app)
        >>> aprefs.timezone = 'Africa/Johannesburg'
        >>> aprefs.dateformat = '%m/%d/%y'
        >>> aprefs.timeformat = '%I:%M %p'
        >>> aprefs.weekstart = calendar.SUNDAY
        >>> request = TestRequest()
        >>> prefs = ViewPreferences(request)
        >>> prefs.timezone.tzname(datetime.utcnow())
        'SAST'
        >>> prefs.timeformat
        '%I:%M %p'
        >>> prefs.dateformat
        '%m/%d/%y'
        >>> prefs.first_day_of_week
        6
    """


def test_suite():
    suite = unittest.TestSuite()
    optionflags = doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS
    suite.addTest(doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                                       optionflags=optionflags))
    suite.addTest(doctest.DocTestSuite('schooltool.app.browser'))
    suite.addTest(doctest.DocFileSuite('../templates.txt'))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Task interfaces
"""

from zope.interface import Interface, Attribute
import zope.schema
from zope.container.interfaces import IContainer, IContained
from zope.container.constraints import contains

from schooltool.person.interfaces import IPerson
from schooltool.common import SchoolToolMessage as _


class ITask(Interface):

    task_id = zope.schema.TextLine(title=_("Task ID"))

    routing_key = zope.schema.TextLine(title=u"Celery routing key",
                                       required=False)

    working = zope.schema.Bool(title=u"Working")
    finished = zope.schema.Bool(title=u"Finished")
    succeeded = zope.schema.Bool(title=_("Succeeded"))
    failed = zope.schema.Bool(title=_("Failed"))
    internal_state = zope.schema.TextLine(title=_("Internal state"))

    permanent_traceback = zope.schema.TextLine(
        title=_("Persisted traceback"),
        required=False)

    permanent_result = Attribute(_("Persisted result"))

    scheduled = zope.schema.Datetime(title=_("Scheduled"))

    creator_username = zope.schema.TextLine(
        title=u"Creator username",
        required=False)

    creator = zope.schema.Object(
        title=u"Creator",
        schema=IPerson, required=False)

    def execute(request):
        """Called remotely."""

    def complete(request, result):
        """Called remotely."""

    def fail(request, result, traceback):
        """Called remotely."""


class IRemoteTask(ITask, IContained):
    pass


class ITaskContainer(IContainer):
    contains(IRemoteTask)


class IMessageBase(Interface):

    sender = zope.schema.Object(
        title=u"Sender",
        schema=Interface,
        required=False)

    sender_id = zope.schema.Int(
        title=u"Sender int id",
        description=u"Int ID of the sender",
        required=False)

    recipients = zope.schema.Set(
        title=u"Recipients",
        description=u"Recipient users or other objects",
        value_type=zope.schema.Object(title=u"Recipient", schema=Interface),
        required=False)

    recipient_ids = zope.schema.Set(
        title=u"Recipient int ids",
        description=u"Int IDs of recipient users or other objects",
        value_type=zope.schema.Int(title=u"Recipient"),
        required=False)

    created_on = zope.schema.Datetime(
       title=u"Created on",
       required=False)

    updated_on = zope.schema.Datetime(
       title=u"Updated on",
       required=False)

    expires_on = zope.schema.Datetime(title=u"Expires on")

    title = zope.schema.TextLine(title=u"Title")

    group = zope.schema.TextLine(title=u"Message group")

    #is_read = zope.schema.Bool(
    #    title=u"Read",
    #    description=u"Message has been read",
    #    )


class IMessage(IMessageBase, IContained):
    """A message sent because of remote task."""


class IProgressMessage(IMessage):
    """Message that has a displayable progress."""


class IMessageContainer(IContainer):
    contains(IMessage)


class ITaskNotification(Interface):

    task = zope.schema.Object(
        title=u"Remote task",
        schema=IRemoteTask,
        required=False)

    request = Attribute('Request')

    def send(*args, **kw):
        """Send out the messages."""


class ITaskScheduledNotification(ITaskNotification):
    pass


class ITaskFailedNotification(ITaskNotification):

    traceback = Attribute('Task traceback')
    result = Attribute('Task result')


class ITaskCompletedNotification(ITaskNotification):

    result = Attribute('Task result')



#
#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Contact information of a Person.

Glue between schooltool.basicperson and schooltool.contact.
"""

from zope.component import getUtility
from zope.component import adapter, adapts
from zope.interface import implements, implementer
from zope.annotation.interfaces import IAnnotations
from zope.security.proxy import removeSecurityProxy
from zope.intid.interfaces import IIntIds
from zope.app.dependable.interfaces import IDependable
from zope.container.interfaces import IContained
from zope.lifecycleevent.interfaces import IObjectAddedEvent
from zope.lifecycleevent.interfaces import IObjectRemovedEvent
from zope.lifecycleevent.interfaces import IObjectModifiedEvent
from zope.lifecycleevent import ObjectAddedEvent, ObjectRemovedEvent
from zope.lifecycleevent import ObjectModifiedEvent
from zope.event import notify

from schooltool.schoolyear.subscriber import ObjectEventAdapterSubscriber
from schooltool.contact.interfaces import IContact
from schooltool.contact.contact import Contact
from schooltool.basicperson.interfaces import IBasicPerson
from schooltool.person.interfaces import IPerson
from schooltool.contact.interfaces import IUniqueFormKey
from schooltool.table.table import simple_form_key


PERSON_CONTACT_KEY = 'schooltool.contact.basicperson'


class IBoundContact(IContact):
    pass


class IBoundContactContained(IBoundContact, IContained):
    pass


class BoundContact(Contact):
    implements(IBoundContactContained)

    @property
    def _person(self):
        person = IBasicPerson(self.__parent__)
        return person

    @property
    def prefix(self):
        return self._person.prefix

    @property
    def first_name(self):
        return self._person.first_name

    @property
    def middle_name(self):
        return self._person.middle_name

    @property
    def last_name(self):
        return self._person.last_name

    @property
    def suffix(self):
        return self._person.suffix

    @property
    def photo(self):
        return self._person.photo

    # XXX: person's language is stored both in a default demographics field
    #      and self contact information now.


@implementer(IPerson)
@adapter(IBoundContact)
def getPersonOfBoundContact(contact):
    return contact.__parent__


@implementer(IContact)
def getBoundContact(context):
    person = IBasicPerson(removeSecurityProxy(context))
    annotations = IAnnotations(person)
    contact = annotations.get(PERSON_CONTACT_KEY, None)
    if contact is None:
        contact = BoundContact()
        annotations[PERSON_CONTACT_KEY] = contact
        contact.__name__ = 'contact'
        contact.__parent__ = person
        dependable = IDependable(contact)
        dependable.addDependent("")
        notify(ObjectAddedEvent(contact, contact.__parent__, contact.__name__))
    return contact


class PersonAddedSubsciber(ObjectEventAdapterSubscriber):
    adapts(IObjectAddedEvent, IBasicPerson)

    def __call__(self):
        # auto-vivification of bound contact
        contact = IContact(self.object)


class PersonModifiedSubsciber(ObjectEventAdapterSubscriber):
    adapts(IObjectModifiedEvent, IBasicPerson)

    def __call__(self):
        contact = IContact(self.object)
        # notify about possible changes in contact information
        # note: we do not provide descriptions of changes
        notify(ObjectModifiedEvent(contact))


class PersonRemovedSubsciber(ObjectEventAdapterSubscriber):
    adapts(IObjectRemovedEvent, IBasicPerson)

    def __call__(self):
        contact = IContact(self.object)
        dependable = IDependable(contact)
        dependable.removeDependent("")
        notify(ObjectRemovedEvent(contact, contact.__parent__, contact.__name__))


@adapter(IBoundContact)
@implementer(IUniqueFormKey)
def getBoundContactFormKey(contact):
    doc_id = getUtility(IIntIds).getId(contact)
    return 'persons.%s%x' % (simple_form_key(contact.__parent__), doc_id)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Scheduling of meetings.
"""
import pytz
import datetime
from persistent import Persistent
from persistent.dict import PersistentDict

from zope.component import getUtility
from zope.container.contained import Contained
from zope.container.btree import BTreeContainer
from zope.intid.interfaces import IIntIds
from zope.interface import implements
from zope.proxy import sameProxiedObjects

from schooltool.common import DateRange
from schooltool.timetable import interfaces


class Period(Persistent, Contained):
    implements(interfaces.IPeriod)

    title = None
    activity_type = None

    def __init__(self, title=None, activity_type=None):
        Contained.__init__(self)
        self.title = title
        self.activity_type = activity_type


class Meeting(object):
    implements(interfaces.IMeeting)

    def __init__(self, dtstart, duration, period=None, meeting_id=None):
        self.dtstart = dtstart
        self.duration = duration
        self.period = period
        self.meeting_id = meeting_id

    def clone(self, **kw):
        dtstart = kw.get('dtstart', self.dtstart)
        duration = kw.get('duration', self.duration)
        period = kw.get('period', self.period)
        meeting_id = kw.get('meeting_id', self.meeting_id)
        return self.__class__(dtstart, duration,
                              period=period, meeting_id=meeting_id)

    def __repr__(self):
        parts = []
        if self.period is not None and self.period.title:
            parts.append(self.period.title)
        if self.dtstart is not None:
            parts.append('on %s' % self.dtstart.strftime('%Y-%m-%d %H:%M %Z'))
        return '<%s%s>' % (
            self.__class__.__name__, parts and ' '+' '.join(parts) or '')


class MeetingException(Persistent, Meeting):
    implements(interfaces.IMeetingException)

    _period_id = None

    @property
    def period(self):
        if self._period_id is None:
            return None
        int_ids = getUtility(IIntIds)
        return int_ids.queryObject(self._period_id)

    @period.setter
    def period(self, value):
        if value is None:
            self._period_id = None
        else:
            int_ids = getUtility(IIntIds)
            self._period_id = int_ids.getId(value)


class Schedule(Contained):
    """A non-persistent abstract schedule."""
    implements(interfaces.ISchedule)

    first = None
    last = None
    title = None
    timezone = None

    def __init__(self, first, last, title=None, timezone='UTC'):
        Contained.__init__(self)
        self.title = title
        self.first = first
        self.last = last
        self.timezone = timezone

    def iterMeetings(date, until_date=None):
        return iter([])


def date_timespan(date, tzinfo=pytz.UTC):
    starts = datetime.datetime.combine(date, datetime.time.min)
    starts = tzinfo.localize(starts)
    ends = datetime.datetime.combine(date, datetime.time.max)
    ends = tzinfo.localize(ends)
    return starts, ends


def iterMeetingsInTimezone(schedule, other_timezone, date, until_date=None):
    if until_date == None:
        until_date = date
    other_timezone = pytz.timezone(other_timezone)
    schedule_timezone = pytz.timezone(schedule.timezone)

    start_time = date_timespan(date, tzinfo=other_timezone)[0]
    end_time = date_timespan(until_date, tzinfo=other_timezone)[1]

    tt_start_date = start_time.astimezone(schedule_timezone).date()
    tt_end_date = end_time.astimezone(schedule_timezone).date()
    if tt_start_date == tt_end_date:
        tt_end_date = None

    meetings = schedule.iterMeetings(tt_start_date, until_date=tt_end_date)
    for meeting in meetings:
        if (meeting.dtstart >= start_time and
            meeting.dtstart <= end_time):
            yield meeting


def iterMeetingsWithExceptions(meetings, exceptions, timezone,
                               date, until_date=None):
    if until_date is None:
        until_date = date

    by_date = {}
    for d in DateRange(date, until_date):
        if d in exceptions:
            by_date[d] = list(exceptions[d])

    tz = pytz.timezone(timezone)

    for original_meeting in meetings:
        meeting_date = original_meeting.dtstart.astimezone(tz).date()

        if meeting_date in exceptions:
            continue

        if meeting_date not in by_date:
            by_date[meeting_date] = list()
        by_date[meeting_date].append(original_meeting)

    for d in sorted(by_date):
        for meeting in sorted(by_date[d], key=lambda m: m.dtstart):
            yield meeting


class ScheduleContainer(BTreeContainer):
    implements(interfaces.IScheduleContainer)

    timezone = None
    exceptions = None

    def __init__(self, timezone='UTC'):
        BTreeContainer.__init__(self)
        self.timezone = timezone
        self.exceptions = PersistentDict()

    @property
    def first(self):
        dates = [schedule.first for schedule in self.values()
                 if schedule.first is not None]
        return dates and min(dates) or None

    @property
    def last(self):
        dates = [schedule.last for schedule in self.values()
                 if schedule.last is not None]
        return dates and max(dates) or None

    def iterOriginalMeetings(self, date, until_date=None):
        if until_date is None:
            until_date = date
        meetings = []
        for schedule in self.values():
            if not sameProxiedObjects(schedule.__parent__, self):
                # We are likely in the process of deleting/moving
                # this schedule.  Ignore.
                continue
            tt_meetings = iterMeetingsInTimezone(
                schedule, self.timezone, date, until_date=until_date)
            meetings.extend(list(tt_meetings))
        return sorted(meetings, key=lambda m: m.dtstart)

    def iterMeetings(self, date, until_date=None):
        meetings = self.iterOriginalMeetings(date, until_date=until_date)
        return iterMeetingsWithExceptions(
            meetings, self.exceptions, self.timezone,
            date, until_date=until_date)
